from autolens.imaging import array_util
import numpy as np
import os

test_data_dir = "{}/../test_files/array/".format(os.path.dirname(os.path.realpath(__file__)))


class TestFits:

    def test__numpy_array_from_fits__3x3_all_ones(self):
        arr = array_util.numpy_array_from_fits(file_path=test_data_dir + '3x3_ones', hdu=0)

        assert (arr == np.ones((3, 3))).all()

    def test__numpy_array_from_fits__4x3_all_ones(self):
        arr = array_util.numpy_array_from_fits(file_path=test_data_dir + '4x3_ones', hdu=0)

        assert (arr == np.ones((4, 3))).all()

    def test__numpy_array_to_fits__output_and_load(self):
        if os.path.exists(test_data_dir + 'test.fits'):
            os.remove(test_data_dir + 'test.fits')

        arr = np.array([[10., 30., 40.],
                        [92., 19., 20.]])

        array_util.numpy_array_to_fits(arr, file_path=test_data_dir + 'test')

        array_load = array_util.numpy_array_from_fits(file_path=test_data_dir + 'test', hdu=0)

        assert (arr == array_load).all()


class TestVariancesFromNoise:

    def test__noise_all_1s__variances_all_1s(self):
        noise = np.array([[1.0, 1.0],
                          [1.0, 1.0]])

        assert (array_util.compute_variances_from_noise(noise) == np.array([[1.0, 1.0],
                                                                            [1.0, 1.0]])).all()

    def test__noise_all_2s__variances_all_4s(self):
        noise = np.array([[2.0, 2.0],
                          [2.0, 2.0]])

        assert (array_util.compute_variances_from_noise(noise) == np.array([[4.0, 4.0],
                                                                            [4.0, 4.0]])).all()

    def test__noise_all_05s__variances_all_025s(self):
        noise = np.array([[0.5, 0.5],
                          [0.5, 0.5]])

        assert (array_util.compute_variances_from_noise(noise) == np.array([[0.25, 0.25],
                                                                            [0.25, 0.25]])).all()


class TestComputeResiduals:

    def test__model_matches_data__residuals_are_all_zero(self):
        data = np.array([[10, 10],
                         [10, 10]])

        model = np.array([[10, 10],
                          [10, 10]])

        result = array_util.compute_residuals(data, model)

        assert result[0, 0] == 0
        assert result[0, 1] == 0
        assert result[1, 0] == 0
        assert result[1, 1] == 0

    def test__model_does_not_match_data__residuals_are_non_zero(self):
        test_image = np.array([[10, 5],
                               [-2, -4.5]])

        model = np.array([[10, 10],
                          [10, -5]])

        result = array_util.compute_residuals(test_image, model)

        assert result[0, 0] == 0  # (10 - 10 = 0)
        assert result[0, 1] == -5  # (5 - 10 = -5)
        assert result[1, 0] == -12  # (-2 - 10 = -12)
        assert result[1, 1] == 0.5  # (-4.5 - (-5)) = 0.5


class TestComputeChiSquared:

    def test__model_matches_data__chi_squareds_all_zeros(self):
        data = np.array([[10, 10],
                         [10, 10]])

        model = np.array([[10, 10],
                          [10, 10]])

        noise = np.array([[1, 1],
                          [1, 1]])

        result = array_util.compute_chi_sq_image(data, model, noise)

        assert result[0, 0] == 0
        assert result[0, 1] == 0
        assert result[1, 0] == 0
        assert result[1, 1] == 0

    def test__model_does_not_match_data__chi_squareds_are_non_zero(self):
        test_image = np.array([[10, 5],
                               [-2, -4.5]])

        model = np.array([[10, 10],
                          [10, -5]])

        noise = np.array([[1, 5],
                          [-1, -2]])

        result = array_util.compute_chi_sq_image(test_image, model, noise)

        assert result[0, 0] == 0  # ( (10 - 10)/1 )^2 = 0
        assert result[0, 1] == 1  # ( (5 - 10)/5 )^2 = ((-)2.5)^2
        assert result[1, 0] == 144  # ( (-2 - 10)/(-1))^2 = 12^2 = 144
        assert result[1, 1] == 0.0625  # ( (-4.5 - (-5))/-2))^2 = (0.5/-3)^2 = (1/6)^2 = 1/16


class TestComputeLikelihood:

    def test__model_matches_data__likelihood_is_zero(self):
        data = np.array([[10, 10],
                         [10, 10]])
        model = np.array([[10, 10],
                          [10, 10]])
        noise = np.array([[1, 1],
                          [1, 1]])

        result = array_util.compute_likelihood(data, model, noise)

        assert result == 0

    def test___model_does_not_match_data__likelihood_computed_correctly(self):
        test_image = np.array([[10, 5],
                               [-2, -4.5]])

        model = np.array([[10, 10],
                          [10, -5]])

        noise = np.array([[1, 5],
                          [-1, -2]])

        result = array_util.compute_likelihood(test_image, model, noise)

        assert result == -72.53125  # -0.5*(0 + 1 + 144 + 0.0625)



import numpy as np
import os
import sys
from matplotlib import pyplot
sys.path.append("../")

from autolens.imaging import imaging, simulate
from astropy.io import fits

def numpy_array_to_fits(array, path, filename):

    new_hdr = fits.Header()
    hdu = fits.PrimaryHDU(array, new_hdr)
    hdu.writeto(path + filename + '.fits')

data_path = "{}/../weighted_data/".format(os.path.dirname(os.path.realpath(__file__)))

pixel_scale = 0.05
dimensions = (256, 256)

psf_dim = (15, 15)

kernel = np.zeros(psf_dim)
sigma = 0.05

for y in range(psf_dim[0]):
    for x in range(psf_dim[1]):

        y0 = (y-((psf_dim[0]/2.0)-0.5))*0.05
        x0 = (x-((psf_dim[1]/2.0)-0.5))*0.05

        kernel[y,x] = np.exp(-(((x0**2)/(2*sigma**2)) + ((y0**2)/(2*sigma**2))))

# psf = imaging.PSF.from_fits(path=data_path+'slacs/', filename='slacs_1_post.fits', hdu=3, pixel_scale=pixel_scale)
# psf.trim(new_dimensions=(15,15))

psf = imaging.PSF(data=kernel, pixel_scale=pixel_scale, renormalize=True)


exposure_time = imaging.ExposureTime.from_one_value(exposure_time=5e3, pixel_dimensions=dimensions,
                                                    pixel_scale=pixel_scale)

background_noise = imaging.NoiseBackground.from_one_value(background_noise=0.000005, pixel_dimensions=dimensions,
                                                          pixel_scale=pixel_scale)

image = simulate.SimulateImage.from_fits(path=data_path, filename='SimLens.fits', hdu=0, pixel_scale=pixel_scale,
                                         exposure_time=exposure_time,
                                         sim_optics=simulate.SimulateOptics(psf=psf),
                                         sim_poisson_noise=simulate.SimulatePoissonNoise(),
                                         sim_background_noise=simulate.SimulateBackgroundNoise(background_noise_sigma=0.005))

noise = imaging.estimate_noise_from_image(image.data, image.exposure_time.data, background_noise.data)
signal_to_noise_ratio = np.divide(image.data, noise)

print(np.max(image.data_original), np.min(image.data_original))
print(np.max(image.data), np.min(image.data))
print(np.max(noise), np.min(noise))
print(np.max(signal_to_noise_ratio))

# pyplot.imshow(psf.weighted_data)
# pyplot.show()
image.plot()

numpy_array_to_fits(image.data, path=data_path, filename='NanSim')

imaging.output_for_fortran(path=data_path, array=image, image_name='NanSim')
imaging.output_for_fortran(path=data_path, array=psf, image_name='NanSim')
imaging.output_for_fortran(path=data_path, array=imaging.Noise(noise, pixel_scale), image_name='NanSim')


import configparser

from autolens import exc
import os
import requests
import zipfile

directory = os.path.dirname(os.path.realpath(__file__))

CONFIG_URL = 'https://drive.google.com/uc?authuser=0&id=1IZE4biWzuxyudDtNr4skyM0PiBHiJhBN&export=download'
CONFIG_DIR = '{}/../'.format(directory)
CONFIG_PATH = '{}/../config'.format(directory)


class NamedConfig(object):
    """Parses generic config"""

    def __init__(self, config_path, section_name):
        """
        Parameters
        ----------
        config_path: String
            The path to the config file
        """
        self.path = config_path
        self.section_name = section_name
        self.parser = configparser.ConfigParser()
        self.parser.read(self.path)

    def get(self, attribute_name, attribute_type=str):
        """

        Parameters
        ----------
        attribute_type: type
            The type to which the value should be cast
        attribute_name: String
            The name of the attribute

        Returns
        -------
        prior_array: []
            An array describing a prior
        """
        string_value = self.parser.get(self.section_name, attribute_name)
        if string_value == "None":
            return None
        if attribute_type is bool:
            return string_value == "True"
        return attribute_type(string_value)

    def has(self, attribute_name):
        """
        Parameters
        ----------
        attribute_name: String
            The name of the attribute

        Returns
        -------
        has_prior: bool
            True iff a prior exists for the module, class and attribute
        """
        return self.parser.has_option(self.section_name, attribute_name)


class AncestorConfig(object):
    """Parses prior config"""

    def __init__(self, config_folder_path):
        """
        Parameters
        ----------
        config_folder_path: String
            The path to the prior config folder
        """
        self.path = config_folder_path
        self.parser = configparser.ConfigParser()

    def read(self, module_name):
        """
        Read a particular config file

        Parameters
        ----------
        module_name: String
            The name of the module for which a config is to be read (priors relate one to one with configs).
        """
        self.parser.read("{}/{}.ini".format(self.path, module_name.split(".")[-1]))

    def get_for_nearest_ancestor(self, cls, attribute_name):
        """
        Find a prior with the attribute name from the config for this class or one of its ancestors

        Parameters
        ----------
        cls: class
            The class of interest
        attribute_name: String
            The name of the attribute
        Returns
        -------
        prior_array: []
            An array describing this prior
        """

        def family(current_class):
            yield current_class
            for next_class in current_class.__bases__:
                for val in family(next_class):
                    yield val

        for family_cls in family(cls):
            if self.has(family_cls.__module__, family_cls.__name__, attribute_name):
                return self.get(family_cls.__module__, family_cls.__name__, attribute_name)

        ini_filename = cls.__module__.split(".")[-1]
        raise exc.PriorException(
            "The prior config at {}/{} does not contain {} in {} or any of its parents".format(self.path,
                                                                                               ini_filename,
                                                                                               attribute_name,
                                                                                               cls.__name__
                                                                                               ))

    def get(self, module_name, class_name, attribute_name):
        """

        Parameters
        ----------
        module_name: String
            The name of the module
        class_name: String
            The name of the class
        attribute_name: String
            The name of the attribute

        Returns
        -------
        prior_array: []
            An array describing a prior
        """
        self.read(module_name)
        return self.parser.get(class_name, attribute_name)

    def has(self, module_name, class_name, attribute_name):
        """
        Parameters
        ----------
        module_name: String
            The name of the module
        class_name: String
            The name of the class
        attribute_name: String
            The name of the attribute

        Returns
        -------
        has_prior: bool
            True iff a prior exists for the module, class and attribute
        """
        self.read(module_name)
        return self.parser.has_option(class_name, attribute_name)


class DefaultPriorConfig(AncestorConfig):
    """Parses prior config"""

    def get(self, module_name, class_name, attribute_name):
        """

        Parameters
        ----------
        module_name: String
            The name of the module
        class_name: String
            The name of the class
        attribute_name: String
            The name of the attribute

        Returns
        -------
        prior_array: []
            An array describing a prior
        """
        arr = super(DefaultPriorConfig, self).get(module_name, class_name, attribute_name).replace(" ", "").split(",")
        return [arr[0]] + list(map(float, arr[1:]))


class WidthConfig(AncestorConfig):
    def get(self, module_name, class_name, attribute_name):
        """

        Parameters
        ----------
        module_name: String
            The name of the module
        class_name: String
            The name of the class
        attribute_name: String
            The name of the attribute

        Returns
        -------
        width: float
            The default width of a gaussian prior for this attribute
        """
        return float(super(WidthConfig, self).get(module_name, class_name, attribute_name))


def is_config():
    return os.path.isdir(CONFIG_PATH)


def download_config():
    zip_path = "{}.zip".format(CONFIG_PATH)
    with requests.get(CONFIG_URL) as response:
        with open(zip_path, 'wb') as f:
            f.write(response.content)

    with zipfile.ZipFile(zip_path, 'r') as z:
        z.extractall(CONFIG_DIR)

    os.remove(zip_path)


if not is_config():
    download_config()



from __future__ import division, print_function

import numpy as np
import pytest
from autolens.profiles import geometry_profiles


class TestEllipticalProfile(object):

    class TestAnglesFromXAxis(object):

        def test__profile_angle_phi_is_0__cosine_and_sin_of_phi_is_1_and_0(self):
            elliptical_profile = geometry_profiles.EllipticalProfile(centre=(1, 1), axis_ratio=1.0, phi=0.0)

            cos_phi, sin_phi = elliptical_profile.cos_and_sin_from_x_axis()

            assert cos_phi == 1.0
            assert sin_phi == 0.0

        def test__profile_angle_phi_is_45__cosine_and_sin_of_phi_follow_trig__therefore_half_root_2(self):
            elliptical_profile = geometry_profiles.EllipticalProfile(centre=(1, 1), axis_ratio=1.0, phi=45.0)

            cos_phi, sin_phi = elliptical_profile.cos_and_sin_from_x_axis()

            assert cos_phi == pytest.approx(0.707, 1e-3)
            assert sin_phi == pytest.approx(0.707, 1e-3)

        def test__profile_angle_phi_is_60__cosine_and_sin_of_phi_follow_trig(self):
            elliptical_profile = geometry_profiles.EllipticalProfile(centre=(1, 1), axis_ratio=1.0, phi=60.0)

            cos_phi, sin_phi = elliptical_profile.cos_and_sin_from_x_axis()

            assert cos_phi == pytest.approx(0.5, 1e-3)
            assert sin_phi == pytest.approx(0.866, 1e-3)

        def test__profile_angle_phi_is_225__cosine_and_sin_of_phi_follow_trig__therefore_negative_half_root_2(self):
            elliptical_profile = geometry_profiles.EllipticalProfile(centre=(1, 1), axis_ratio=1.0, phi=225.0)

            cos_phi, sin_phi = elliptical_profile.cos_and_sin_from_x_axis()

            assert cos_phi == pytest.approx(-0.707, 1e-3)
            assert sin_phi == pytest.approx(-0.707, 1e-3)

    class TestTransformToEllipticalReferenceFrame(object):

        def test__profile_angle_phi_is_0__grid_x_1_y_1__returns_same_grid_so_x_1_y_1(self):

            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=0.0)

            grid = np.array([[1.0, 1.0]])

            transformed_grid = elliptical_profile.transform_grid_to_reference_frame_jitted(grid)

            assert transformed_grid[0,0] == pytest.approx(1.0, 1e-3)
            assert transformed_grid[0,1] == pytest.approx(1.0, 1e-3)

        def test__profile_angle_phi_90__grid_x_1_y_1__rotated_grid_clockwise_so_x_1_y_negative_1(self):

            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=90.0)

            grid = np.array([[1.0, 1.0]])

            transformed_grid = elliptical_profile.transform_grid_to_reference_frame_jitted(grid)

            assert transformed_grid[0,0] == pytest.approx(1.0, 1e-3)
            assert transformed_grid[0,1] == pytest.approx(-1.0, 1e-3)

        def test__profile_angle_phi_180__grid_x_1_y_1__rotated_grid_clockwise_so_x_and_y_negative_1(self):
            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=180.0)

            grid = np.array([[1.0, 1.0]])

            transformed_grid = elliptical_profile.transform_grid_to_reference_frame_jitted(grid)

            assert transformed_grid[0,0] == pytest.approx(-1.0, 1e-3)
            assert transformed_grid[0,1] == pytest.approx(-1.0, 1e-3)

        def test__profile_angle_phi_270__grid_x_1_y_1__rotated_grid_clockwise_so_x_negative_1_y_1(self):

            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=270.0)

            grid = np.array([[1.0, 1.0]])

            transformed_grid = elliptical_profile.transform_grid_to_reference_frame_jitted(grid)

            assert transformed_grid[0,0] == pytest.approx(-1.0, 1e-3)
            assert transformed_grid[0,1] == pytest.approx(1.0, 1e-3)

        def test__profile_angle_phi_360__rotated_grid_are_original_grid_x_1_y_(self):

            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=360.0)

            grid = np.array([[1.0, 1.0]])

            transformed_grid = elliptical_profile.transform_grid_to_reference_frame_jitted(grid)

            assert transformed_grid[0,0] == pytest.approx(1.0, 1e-3)
            assert transformed_grid[0,1] == pytest.approx(1.0, 1e-3)

        def test__profile_angle_phi_315__grid_x_1_y_1__rotated_grid_clockwise_so_x_0_y_root_2(self):

            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=315.0)

            grid = np.array([[1.0, 1.0]])

            transformed_grid = elliptical_profile.transform_grid_to_reference_frame_jitted(grid)

            assert transformed_grid[0,0] == pytest.approx(0.0, 1e-3)
            assert transformed_grid[0,1] == pytest.approx(2 ** 0.5, 1e-3)

        def test__include_profile_centre_offset__is_used_before_rotation_is_performed(self):

            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=90.0, centre=(2.0, 3.0))

            grid = np.array([[3.0, 4.0]])

            transformed_grid = elliptical_profile.transform_grid_to_reference_frame_jitted(grid)

            assert transformed_grid[0,0] == pytest.approx(1.0, 1e-3)
            assert transformed_grid[0,1] == pytest.approx(-1.0, 1e-3)

    class TestTransformCoordinatesBackToCartesian(object):

        def test___profile_angle_phi_0__therefore_no_rotation_and_grid_are_unchanged(self):
            
            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=0.0)

            transformed_grid = elliptical_profile.transform_grid_from_reference_frame(np.array([[1.0, 1.0]]))

            assert transformed_grid[0,0] == 1.0
            assert transformed_grid[0,1] == 1.0

        def test__profile_angle_phi_90__grid_x_1_y_1__rotated_counter_clockwise_to_x_negative_1_y_1(self):

            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=90.0)

            transformed_grid = elliptical_profile.transform_grid_from_reference_frame(np.array([[1.0, 1.0]]))

            assert transformed_grid[0,0] == pytest.approx(-1.0, 1e-3)
            assert transformed_grid[0,1] == 1.0

        def test___profile_angle_phi_45__grid_x_1_y_1__rotated_counter_clockwise_to_x_0_y_root_2(self):

            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=45.0)

            transformed_grid = elliptical_profile.transform_grid_from_reference_frame(np.array([[1.0, 1.0]]))

            assert transformed_grid[0,0] == pytest.approx(0.0, 1e-3)
            assert transformed_grid[0,1] == pytest.approx(2 ** 0.5, 1e-3)

    class TestCoordinateMovements(object):

        def test__grid_and_centres_of_two_lenses_are_equivalent__grid_are_equivalent(self):
            
            elliptical_profile1 = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=0.0, centre=(0, 0))
            grid1 = np.array([[1.0, 1.0]])
            grid1 = elliptical_profile1.transform_grid_to_reference_frame_jitted(grid1)

            elliptical_profile2 = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=0.0, centre=(-1, -1))
            grid2 = np.array([[0.0, 0.0]])
            grid2 = elliptical_profile2.transform_grid_to_reference_frame_jitted(grid2)

            assert grid1[0,0] == grid2[0,0]
            assert grid1[0,1] == grid2[0,1]

        def test__same_as_above_but_include_angle_phi_as_55__grid_are_equivalent(self):
            elliptical_profile1 = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=55.0, centre=(0, 0))
            grid1 = np.array([[1.0, 1.0]])
            grid1 = elliptical_profile1.transform_grid_to_reference_frame_jitted(grid1)

            elliptical_profile2 = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=55.0, centre=(-1, -1))
            grid2 = np.array([[0.0, 0.0]])
            grid2 = elliptical_profile2.transform_grid_to_reference_frame_jitted(grid2)

            assert grid1[0,0] == grid2[0,0]
            assert grid1[0,1] == grid2[0,1]

        def test__grid_are_again_the_same_after_centre_shift__grid_equivalent(self):
            elliptical_profile1 = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=55.0, centre=(1, 1))
            grid1 = np.array([[1.0, 1.0]])
            grid1 = elliptical_profile1.transform_grid_to_reference_frame_jitted(grid1)

            elliptical_profile2 = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=55.0, centre=(-1, -1))
            grid2 = np.array([[-1.0, -1.0]])
            grid2 = elliptical_profile2.transform_grid_to_reference_frame_jitted(grid2)

            assert grid1[0,0] == grid2[0,0]
            assert grid1[0,1] == grid2[0,1]

    class TestRotateCoordinatesThenBackToCartesian(object):

        def test__grid_are_transformed_to_and_from_reference_frame__equal_to_original_values(self):
            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=315.0)

            grid_original = np.array([[5.2221, 2.6565]])

            grid_elliptical = elliptical_profile.transform_grid_to_reference_frame_jitted(grid_original)

            transformed_grid = elliptical_profile.transform_grid_from_reference_frame(grid_elliptical)

            assert transformed_grid[0,0] == pytest.approx(grid_original[0,0], 1e-5)
            assert transformed_grid[0,1] == pytest.approx(grid_original[0,1], 1e-5)

        def test__same_as_above_different_values(self):

            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=160.232)

            grid_original = np.array([[3.2, -76.6]])

            grid_elliptical = elliptical_profile.transform_grid_to_reference_frame_jitted(grid_original)

            transformed_grid = elliptical_profile.transform_grid_from_reference_frame(grid_elliptical)

            assert transformed_grid[0,0] == pytest.approx(grid_original[0,0], 1e-2)
            assert transformed_grid[0,1] == pytest.approx(grid_original[0,1], 1e-2)

        def test__same_again_another_set_of_values(self):
            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=174.342)

            grid_original = np.array([[-42.2, -93.6]])

            grid_elliptical = elliptical_profile.transform_grid_to_reference_frame_jitted(grid_original)

            transformed_grid = elliptical_profile.transform_grid_from_reference_frame(grid_elliptical)

            assert transformed_grid[0,0] == pytest.approx(grid_original[0,0], 1e-2)
            assert transformed_grid[0,1] == pytest.approx(grid_original[0,1], 1e-2)

    class TestTransformedGridToEccentricRadius(object):

        def test__profile_axis_ratio_1__r_is_root_2__therefore_ecccentric_radius_is_elliptical_radius_is_root_2(self):

            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1.0, phi=0.0)

            eccentric_radius = elliptical_profile.grid_to_eccentric_radii(np.array([[1.0, 1.0]]))

            assert eccentric_radius == pytest.approx(2.0 ** 0.5, 1e-3)

        def test__same_as_above_but_axis_ratio_is_05__r_follows_elliptical_coordinate_equation(self):
            
            # eccentric_radius = sqrt(axis_ratio) * sqrt(  x**2 +   (y**2 / axis_ratio**2))
            # eccentric_radius =        sqrt(0.5) * sqrt(1.0**2 + (1.0**2 /        0.5**2))
            # eccentric radius =        sqrt(0.5) * sqrt( 5 ) = 1.58113

            elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=0.5, phi=0.0)

            eccentric_radius = elliptical_profile.grid_to_eccentric_radii(np.array([[1.0, 1.0]]))

            assert eccentric_radius == pytest.approx(1.58113, 1e-3)


class TestSphericalProfile(object):

    class TestCoordinatesMovement(object):

        def test__profile_cenre_x_0_y_0__grid_x_1_y_1__no_coordinate_movement_so_x_1_y_1(self):

            spherical_profile = geometry_profiles.SphericalProfile(centre=(0.0, 0.0))

            transformed_grid = spherical_profile.transform_grid_from_reference_frame(np.array([[1.0, 1.0]]))

            assert transformed_grid[0,0] == 1.0
            assert transformed_grid[0,1] == 1.0

        def test__grid_and_centres_of_two_lenses_are_equivalent__grid_are_equivalent(self):

            spherical_profile1 = geometry_profiles.SphericalProfile(centre=(0, 0))

            grid1 = np.array([[1.0, 1.0]])
            grid1 = spherical_profile1.transform_grid_to_reference_frame_jitted(grid1)

            spherical_profile2 = geometry_profiles.SphericalProfile(centre=(-1, -1))
            grid2 = np.array([[0.0, 0.0]])
            grid2 = spherical_profile2.transform_grid_to_reference_frame_jitted(grid2)

            assert grid1[0,0] == grid2[0,0]
            assert grid1[0,1] == grid2[0,1]

        def test__grid_are_again_the_same_after_centre_shift__grid_equivalent(self):

            spherical_profile1 = geometry_profiles.SphericalProfile(centre=(1, 1))
            grid1 = np.array([[1.0, 1.0]])
            grid1 = spherical_profile1.transform_grid_to_reference_frame_jitted(grid1)

            spherical_profile2 = geometry_profiles.SphericalProfile(centre=(-1, -1))
            grid2 = np.array([[-1.0, -1.0]])
            grid2 = spherical_profile2.transform_grid_to_reference_frame_jitted(grid2)

            assert grid1[0,0] == grid2[0,0]
            assert grid1[0,1] == grid2[0,1]

    class TestTransformCoordinatesAndThenBackToCartesian(object):

        def test__profile_centre_x_0_y_0__grid_x_1_y_1__returns_x_1_y_1_so_same_grid(self):

            spherical_profile = geometry_profiles.SphericalProfile(centre=(0.0, 0.0))

            grid = np.array([[1.0, 1.0]])

            transformed_grid = spherical_profile.transform_grid_to_reference_frame_jitted(grid)

            assert transformed_grid[0,0] == pytest.approx(1.0, 1e-3)
            assert transformed_grid[0,1] == pytest.approx(1.0, 1e-3)

        def test__grid_are_transformed_to_and_from_reference_frame__equal_to_original_values(self):
            spherical_profile = geometry_profiles.SphericalProfile(centre=(0.0, 0.0))

            grid_original = np.array([[5.2221, 2.6565]])

            grid_spherical = spherical_profile.transform_grid_to_reference_frame_jitted(grid_original)

            transformed_grid = spherical_profile.transform_grid_from_reference_frame(grid_spherical)

            assert transformed_grid[0,0] == pytest.approx(grid_original[0,0], 1e-5)
            assert transformed_grid[0,1] == pytest.approx(grid_original[0,1], 1e-5)

        def test__same_as_above_different_grid__equal_to_original_value(self):

            spherical_profile = geometry_profiles.SphericalProfile(centre=(0.0, 0.0))

            grid_original = np.array([[3.2, -76.6]])

            grid_spherical = spherical_profile.transform_grid_to_reference_frame_jitted(grid_original)

            transformed_grid = spherical_profile.transform_grid_from_reference_frame(grid_spherical)

            assert transformed_grid[0,0] == pytest.approx(grid_original[0,0], 1e-2)
            assert transformed_grid[0,1] == pytest.approx(grid_original[0,1], 1e-2)


# class TestTransform(object):
#
#     def test_symmetry(self):
#         p = geometry_profiles.EllipticalProfile((3, 5), 2, 2)
#         assert (p.transform_grid_from_reference_frame(p.transform_grid_to_reference_frame((5, 7))) == (5, 7)).all()


class TestFromProfile(object):

    def test__profile_from_profile__centre_x_1_y_1_is_passed(self):
        p = geometry_profiles.Profile(centre=(1, 1))
        assert geometry_profiles.Profile.from_profile(p).centre == (1, 1)

    def test__elliptical_profile_from_profile__centre_x_1_y_1__axis_ratio_1__phi_2__all_are_passed(self):
        p = geometry_profiles.Profile(centre=(1, 1))
        elliptical_profile = geometry_profiles.EllipticalProfile.from_profile(p, axis_ratio=1, phi=2)
        assert elliptical_profile.__class__ == geometry_profiles.EllipticalProfile
        assert elliptical_profile.centre == (1, 1)
        assert elliptical_profile.axis_ratio == 1
        assert elliptical_profile.phi == 2

    def test__profile_from_elliptical_profile__centre_x_1_y_1_is_passed(self):
        elliptical_profile = geometry_profiles.EllipticalProfile(centre=(1, 1), axis_ratio=1, phi=2)
        p = geometry_profiles.Profile.from_profile(elliptical_profile)
        assert p.__class__ == geometry_profiles.Profile
        assert p.centre == (1, 1)

    def test__elliptcal_profile_from_elliptical_profile__optionally_override_axis_ratio_with_3(self):
        elliptical_profile = geometry_profiles.EllipticalProfile(axis_ratio=1, phi=2)
        new_geometry_profile = geometry_profiles.EllipticalProfile.from_profile(elliptical_profile, axis_ratio=3)
        assert new_geometry_profile.phi == 2
        assert new_geometry_profile.axis_ratio == 3




from autolens.pipeline import phase as ph
import pytest
from autolens.analysis import galaxy as g
from autolens.analysis import galaxy_prior as gp
from autolens.autopipe import non_linear
import numpy as np
from autolens.imaging import mask as msk
from autolens.imaging import image as img
from autolens.imaging import masked_image as mi
from autolens.autopipe import model_mapper as mm
from autolens.profiles import light_profiles

shape = (10, 10)


class MockResults(object):
    def __init__(self, model_image, galaxy_images=()):
        self.model_image = model_image
        self.galaxy_images = galaxy_images
        self.constant = mm.ModelInstance()
        self.variable = mm.ModelMapper()


class NLO(non_linear.NonLinearOptimizer):
    def fit(self, analysis):
        class Fitness(object):
            def __init__(self, instance_from_physical_vector, constant):
                self.result = None
                self.instance_from_physical_vector = instance_from_physical_vector
                self.constant = constant

            def __call__(self, vector):
                instance = self.instance_from_physical_vector(vector)
                for key, value in self.constant.__dict__.items():
                    setattr(instance, key, value)

                likelihood = analysis.fit(**instance.__dict__)
                self.result = non_linear.Result(instance, likelihood)

                # Return Chi squared
                return -2 * likelihood

        fitness_function = Fitness(self.variable.instance_from_physical_vector, self.constant)
        fitness_function(self.variable.total_parameters * [0.5])

        return fitness_function.result


@pytest.fixture(name="grids")
def make_grids(masked_image):
    return msk.GridCollection.from_mask_sub_grid_size_and_blurring_shape(
        masked_image.mask, 1, masked_image.psf.shape)


@pytest.fixture(name="phase")
def make_phase():
    return ph.SourceLensPhase(optimizer_class=NLO)


@pytest.fixture(name="galaxy")
def make_galaxy():
    return g.Galaxy()


@pytest.fixture(name="galaxy_prior")
def make_galaxy_prior():
    return gp.GalaxyPrior()


@pytest.fixture(name="image")
def make_image():
    image = img.Image(np.array(np.zeros(shape)), pixel_scale=1.0, psf=img.PSF(np.ones((3, 3))), noise=np.ones(shape))
    return image


@pytest.fixture(name="masked_image")
def make_masked_image():
    image = img.Image(np.array(np.zeros(shape)), pixel_scale=1.0, psf=img.PSF(np.ones((3, 3))), noise=np.ones(shape))
    mask = msk.Mask.circular(shape, 1, 3)
    return mi.MaskedImage(image, mask)


@pytest.fixture(name="results")
def make_results():
    return MockResults(np.ones((10, 10)),
                       galaxy_images=[np.ones((10, 10)), np.ones((10, 10))])


@pytest.fixture(name="results_collection")
def make_results_collection(results):
    return ph.ResultsCollection([results])


class TestPhase(object):

    def test_set_constants(self, phase, galaxy):
        phase.lens_galaxy = galaxy
        assert phase.optimizer.constant.lens_galaxy == galaxy
        assert not hasattr(phase.optimizer.variable, "lens_galaxy")

    def test_set_variables(self, phase, galaxy_prior):
        phase.lens_galaxy = galaxy_prior
        assert phase.optimizer.variable.lens_galaxy == galaxy_prior
        assert not hasattr(phase.optimizer.constant, "lens_galaxy")

    def test_mask_analysis(self, phase, image, masked_image):
        analysis = phase.make_analysis(image=image)
        assert analysis.last_results is None
        assert analysis.masked_image == masked_image

    def test_fit(self, phase, image):
        phase.source_galaxy = g.Galaxy()
        phase.lens_galaxy = g.Galaxy()
        result = phase.run(image=image)
        assert isinstance(result.constant.lens_galaxy, g.Galaxy)
        assert isinstance(result.constant.source_galaxy, g.Galaxy)

    def test_customize(self, results, image):
        class MyPhase(ph.SourceLensPhase):
            def pass_priors(self, previous_results):
                self.lens_galaxy = previous_results.last.constant.lens_galaxy
                self.source_galaxy = previous_results.last.variable.source_galaxy

        galaxy = g.Galaxy()
        galaxy_prior = gp.GalaxyPrior()

        setattr(results.constant, "lens_galaxy", galaxy)
        setattr(results.variable, "source_galaxy", galaxy_prior)

        phase = MyPhase(optimizer_class=NLO)
        phase.make_analysis(image=image, previous_results=ph.ResultsCollection([results]))

        assert phase.lens_galaxy == galaxy
        assert phase.source_galaxy == galaxy_prior

    def test_phase_property(self):

        class MyPhase(ph.SourceLensPhase):
            prop = ph.phase_property("prop")

        phase = MyPhase(optimizer_class=NLO)

        phase.prop = gp.GalaxyPrior()

        assert phase.variable.prop == phase.prop

        galaxy = g.Galaxy()
        phase.prop = galaxy

        assert phase.constant.prop == galaxy
        assert not hasattr(phase.variable, "prop")

        phase.prop = gp.GalaxyPrior()
        assert not hasattr(phase.constant, "prop")

    def test_default_mask_function(self, phase, image):
        assert len(mi.MaskedImage(image, phase.mask_function(image))) == 32

    def test_galaxy_images(self, image, phase):
        phase.lens_galaxy = g.Galaxy()
        phase.source_galaxy = g.Galaxy()
        result = phase.run(image)
        assert len(result.galaxy_images) == 2

    def test_duplication(self):
        phase = ph.SourceLensPhase(lens_galaxy=gp.GalaxyPrior(), source_galaxy=gp.GalaxyPrior())

        ph.SourceLensPhase()

        assert phase.lens_galaxy is not None
        assert phase.source_galaxy is not None

    def test_modify_image(self, image):
        class MyPhase(ph.Phase):
            def modify_image(self, im, previous_results):
                assert image.shape == im.shape
                return im

        phase = MyPhase()
        analysis = phase.make_analysis(image)
        assert analysis.masked_image != image

    def test_model_images(self, image):
        phase = ph.SourceLensPhase()
        analysis = phase.make_analysis(image)
        instance = mm.ModelInstance()
        instance.lens_galaxy = g.Galaxy(light=light_profiles.EllipticalExponential())
        instance.source_galaxy = None

        images = analysis.galaxy_images_for_model(instance)
        assert images[0].shape == image.shape
        assert images[1] is None


class TestPixelizedPhase(object):
    def test_constructor(self):
        phase = ph.PixelizedSourceLensPhase()
        assert isinstance(phase.source_galaxy, gp.GalaxyPrior)
        assert phase.lens_galaxy is None


class TestAnalysis(object):
    def test_model_image(self, results_collection, masked_image):
        analysis = ph.Phase.Analysis(results_collection, masked_image)
        assert (results_collection[0].model_image == analysis.last_results.model_image).all()


class TestResult(object):
    def test_model_image(self):
        result = ph.Phase.Result(mm.ModelInstance(), 1, mm.ModelMapper(), [np.array([1, 2, 3]), np.array([2, 3, 4])])
        assert (result.model_image == np.array([3, 5, 7])).all()

    def test_results(self):
        results = ph.ResultsCollection([1, 2, 3])
        assert results == [1, 2, 3]
        assert results.last == 3
        assert results.first == 1



import numpy as np

class RegularizationConstant(object):

    pixels = None
    regularization_coefficients = None

    def regularization_matrix_from_pix_neighbors(self, pix_neighbors):
        """
        Setup a pixelization's constant regularization matrix (see test_pixelization.py)

        Parameters
        ----------
        pix_neighbors : [[]]
            A list of the neighbors of each pixel.
        """

        regularization_matrix = np.zeros(shape=(self.pixels, self.pixels))

        reg_coeff = self.regularization_coefficients[0] ** 2.0

        for i in range(self.pixels):
            regularization_matrix[i, i] += 1e-8
            for j in pix_neighbors[i]:
                regularization_matrix[i, i] += reg_coeff
                regularization_matrix[i, j] -= reg_coeff

        return regularization_matrix


class RegularizationWeighted(object):

    pixels = None
    regularization_coefficients = None
    pix_signal_scale = None

    def pix_signals_from_images(self, image_to_pix, galaxy_image):
        """Compute the (scaled) signal in each pixel, where the signal is the sum of its image-pixel fluxes. \
        These pix-signals are then used to compute the effective regularization weight of each pixel.

        The pix signals are scaled in the following ways:

        1) Divided by the number of image-pixels in the pixel, to ensure all pixels have the same \
        'relative' signal (i.e. a pixel with 10 images-pixels doesn't have x2 the signal of one with 5).

        2) Divided by the maximum pix-signal, so that all signals vary between 0 and 1. This ensures that the \
        regularizations weights they're used to compute are defined identically for all image units / SNR's.

        3) Raised to the power of the hyper-parameter *pix_signal_scale*, so the method can control the relative \
        contribution of the different regions of regularization.
        """

        pix_signals = np.zeros((self.pixels,))
        pix_sizes = np.zeros((self.pixels,))

        for image_index in range(galaxy_image.shape[0]):
            pix_signals[image_to_pix[image_index]] += galaxy_image[image_index]
            pix_sizes[image_to_pix[image_index]] += 1

        pix_signals /= pix_sizes
        pix_signals /= max(pix_signals)

        return pix_signals ** self.pix_signal_scale

    def regularization_weights_from_pix_signals(self, pix_signals):
        """Compute the regularization weights, which represent the effective regularization coefficient of every \
        pixel. These are computed using the (scaled) pix-signal in each pixel.

        Two regularization coefficients are used which map to:

        1) pix_signals - This regularizes pix-plane pixels with a high pix-signal (i.e. where the pix is).
        2) 1.0 - pix_signals - This regularizes pix-plane pixels with a low pix-signal (i.e. background sky)
        """
        return (self.regularization_coefficients[0] * pix_signals +
                self.regularization_coefficients[1] * (1.0 - pix_signals)) ** 2.0

    def regularization_matrix_from_pix_neighbors(self, regularization_weights, pix_neighbors):
        """
        Setup a weighted regularization matrix, where all pixels are regularized with one another in both \
        directions using a different effective regularization coefficient.

        Parameters
        ----------
        regularization_weights : list(float)
            The regularization weight of each pixel
        pix_neighbors : [[]]
            A list of the neighbors of each pixel.
        """

        regularization_matrix = np.zeros(shape=(self.pixels, self.pixels))

        reg_weight = regularization_weights ** 2.0

        for i in range(self.pixels):
            for j in pix_neighbors[i]:
                regularization_matrix[i, i] += reg_weight[j]
                regularization_matrix[j, j] += reg_weight[j]
                regularization_matrix[i, j] -= reg_weight[j]
                regularization_matrix[j, i] -= reg_weight[j]

        return regularization_matrix


from autolens.imaging import image as im
from autolens.imaging import mask as msk
from autolens.imaging import convolution
import numpy as np


class MaskedImage(im.Image):
    def __new__(cls, image, mask, sub_grid_size=2, profile_psf_shape=None, pixelization_psf_shape=None):
        return np.array(mask.map_to_1d(image)).view(cls)

    def __init__(self, image, mask, sub_grid_size=2, profile_psf_shape=None, pixelization_psf_shape=None):
        """
        An image that has been masked. Only data within the mask is kept. This data is kept in 1D with a corresponding
        array mapping data back to 2D.

        Parameters
        ----------
        image: im.Image
            A 2D image
        mask: msk.Mask
            A mask to be applied to the image
        sub_grid_size : int

        """
        super().__init__(array=image, pixel_scale=image.pixel_scale, noise=mask.map_to_1d(image.noise), psf=image.psf)

        self.image = image
        self.image_shape = image.shape
        self.mask = mask
        self.sub_grid_size = sub_grid_size

        if profile_psf_shape is None:
            profile_psf_shape = self.image.psf.shape
        if pixelization_psf_shape is None:
            pixelization_psf_shape = self.image.psf.shape

        self.blurring_mask = mask.blurring_mask_for_kernel_shape(profile_psf_shape)
        self.convolver_image = convolution.ConvolverImage(self.mask,
                                                          self.blurring_mask, self.image.psf.trim(profile_psf_shape))
        self.convolver_mapping_matrix = convolution.ConvolverMappingMatrix(self.mask,
                                                                           self.image.psf.trim(pixelization_psf_shape))
        self.grids = msk.GridCollection.from_mask_sub_grid_size_and_blurring_shape(mask=mask, sub_grid_size=sub_grid_size,
                                                                                   blurring_shape=profile_psf_shape)

        self.borders = msk.BorderCollection.from_mask_and_sub_grid_size(mask=mask, sub_grid_size=sub_grid_size)

    def __array_finalize__(self, obj):
        super(MaskedImage, self).__array_finalize__(obj)
        if isinstance(obj, MaskedImage):
            self.image = obj.image
            self.image_shape = obj.image_shape
            self.mask = obj.mask
            self.blurring_mask = obj.blurring_mask
            self.convolver_image = obj.convolver_image
            self.convolver_mapping_matrix = obj.convolver_mapping_matrix
            self.grids = obj.grids
            self.borders = obj.borders



    def map_to_2d(self, data):
        """Use mapper to map an input data-set from a *GridData* to its original 2D image.
        Parameters
        -----------
        data : ndarray
            The grid-data which is mapped to its 2D image.
        """
        data_2d = np.zeros(self.image_shape)

        for (i, pixel) in enumerate(self.mask.grid_to_pixel()):
            data_2d[pixel[0], pixel[1]] = data[i]

        return data_2d



import numpy as np
import numba


class Reconstructor(object):

    def __init__(self, mapping, regularization, image_to_pix, sub_to_pix):
        """The matrices and mappings used to linearly invert and fit a data-set.

        Parameters
        -----------
        mapping : ndarray
            The matrix representing the mapping between reconstruction-pixels and weighted_data-pixels.
        regularization : ndarray
            The matrix defining how the reconstruction's pixels are regularized with one another when fitting the
            weighted_data.
        image_to_pix : ndarray
            The mapping between each image-grid pixel and pixelization-grid pixel.
        sub_to_pix : ndarray
            The mapping between each sub-grid pixel and pixelization-grid sub-pixel.
        """
        self.mapping = mapping
        self.mapping_shape = mapping.shape
        self.regularization = regularization
        self.image_to_pix = image_to_pix
        self.sub_to_pix = sub_to_pix

    def reconstruct_image(self, image, noise, convolver):
        """Fit the image data using the inversion."""

        blurred_mapping = convolver.convolve_mapping_matrix_jit(self.mapping)
        covariance = self.covariance_matrix_from_blurred_mapping_jit(blurred_mapping, noise)
        data_vector = self.data_vector_from_blurred_mapping_and_data_jit(blurred_mapping, image, noise)
        cov_reg = np.add(covariance, self.regularization)
        reconstruction = np.linalg.solve(cov_reg, data_vector)

        return Reconstruction(data_vector, blurred_mapping, self.regularization, covariance, cov_reg, reconstruction)

    def covariance_matrix_from_blurred_mapping_jit(self, blurred_mapping, noise_vector):
        flist = np.zeros(blurred_mapping.shape[0])
        iflist = np.zeros(blurred_mapping.shape[0], dtype='int')
        return self.covariance_matrix_from_blurred_mapping_jitted(blurred_mapping, noise_vector, flist, iflist)

    @staticmethod
    @numba.jit(nopython=True)
    def covariance_matrix_from_blurred_mapping_jitted(blurred_mapping, noise_vector, flist, iflist):

        mapping_shape = blurred_mapping.shape

        covariance_matrix = np.zeros((mapping_shape[1], mapping_shape[1]))

        for image_index in range(mapping_shape[0]):
            index=0
            for pix_index in range(mapping_shape[1]):
                if blurred_mapping[image_index, pix_index] > 0.0:
                    index += 1
                    flist[index] = blurred_mapping[image_index, pix_index] / noise_vector[image_index]
                    iflist[index] = pix_index

            if index > 0:
                for i1 in range(index+1):
                    for j1 in range(index+1):
                        ix = iflist[i1]
                        iy = iflist[j1]
                        covariance_matrix[ix, iy] += flist[i1]*flist[j1]

        for i in range(mapping_shape[1]):
            for j in range(mapping_shape[1]):
                covariance_matrix[i, j] = covariance_matrix[j, i]

        return covariance_matrix

    def data_vector_from_blurred_mapping_and_data_jit(self, blurred_mapping, image_vector, noise_vector):
        """ Compute the covariance matrix directly - used to integration test that our covariance matrix generator approach
        truly works."""
        return self.data_vector_from_blurred_mapping_and_data_jitted(blurred_mapping, image_vector, noise_vector)

    @staticmethod
    @numba.jit(nopython=True)
    def data_vector_from_blurred_mapping_and_data_jitted(blurred_mapping, image_vector, noise_vector):
        """ Compute the covariance matrix directly - used to integration test that our covariance matrix generator approach
        truly works."""

        mapping_shape = blurred_mapping.shape

        data_vector = np.zeros(mapping_shape[1])

        for image_index in range(mapping_shape[0]):
            for pix_index in range(mapping_shape[1]):
                data_vector[pix_index] += image_vector[image_index] * \
                                          blurred_mapping[image_index, pix_index] / (noise_vector[image_index] ** 2.0)

        return data_vector


class Reconstruction(object):

    def __init__(self, data_vector, blurred_mapping, regularization, covariance, covariance_regularization,
                 reconstruction):
        """The matrices, mappings which have been used to linearly invert and fit a data-set.

        Parameters
        -----------
        data_vector : ndarray | None
            The 1D vector representing the data, weighted by its noise in a chi squared sense, which is fitted by the \
            inversion (D).
        blurred_mapping : ndarray | None
            The matrix representing the mapping between reconstruction-pixels and data-pixels, including a \
            blurring operation (f).
        regularization : ndarray | None
            The matrix defining how the reconstruction's pixels are regularized with one another (H).
        covariance : ndarray | None
            The covariance between each reconstruction pixel and all other reconstruction pixels (F).
        covariance_regularization : ndarray | None
            The covariance + regularizationo matrix.
        reconstruction : ndarray | None
            The vector containing the reconstructed fit of the data.
        """
        self.weighted_data = data_vector
        self.blurred_mapping = blurred_mapping
        self.regularization = regularization
        self.covariance = covariance
        self.covariance_regularization = covariance_regularization
        self.reconstruction = reconstruction

    def model_image_from_reconstruction_jit(self):
        """ Map the reconstruction pix s_vector back to the image-plane to compute the pixelization's model-image.
        """
        return self.model_image_from_reconstruction_jitted(self.reconstruction, self.blurred_mapping)
    
    @staticmethod
    @numba.jit(nopython=True)
    def model_image_from_reconstruction_jitted(reconstruction, blurred_mapping):
        """ Map the reconstruction pix s_vector back to the image-plane to compute the pixelization's model-image.
        """
        model_image = np.zeros(blurred_mapping.shape[0])
        for i in range(blurred_mapping.shape[0]):
            for j in range(reconstruction.shape[0]):
                model_image[i] += reconstruction[j] * blurred_mapping[i, j]

        return model_image

    def regularization_term_from_reconstruction(self):
        """ Compute the regularization term of a pixelization's Bayesian likelihood function. This represents the sum \
         of the difference in fluxes between every pair of neighboring pixels. This is computed as:

         s_T * H * s = s_vector.T * regularization_matrix_const * s_vector

         The term is referred to as 'G_l' in Warren & Dye 2003, Nightingale & Dye 2015.

         The above works include the regularization coefficient (lambda) in this calculation. In PyAutoLens, this is  \
         already in the regularization matrix and thus included in the matrix multiplication.
         """
        return np.matmul(self.reconstruction.T, np.matmul(self.regularization, self.reconstruction))

    @staticmethod
    def log_determinant_of_matrix_cholesky(matrix):
        """There are two terms in the pixelization's Bayesian likelihood function which require the log determinant of \
        a matrix. These are (Nightingale & Dye 2015, Nightingale, Dye and Massey 2018):

        ln[det(F + H)] = ln[det(cov_reg_matrix)]
        ln[det(H)]     = ln[det(regularization_matrix_const)]

        The cov_reg_matrix is positive-definite, which means its log_determinant can be computed efficiently \
        (compared to using np.det) by using a Cholesky decomposition first and summing the log of each diagonal term.

        Parameters
        -----------
        matrix : ndarray
            The positive-definite matrix the log determinant is computed for.
        """
        return 2.0 * np.sum(np.log(np.diag(np.linalg.cholesky(matrix))))


from autolens.analysis import galaxy_prior
from autolens.autopipe import non_linear
from autolens.profiles import light_profiles, mass_profiles


# Test should work once file_param_names has been fixed
def _test_pipeline():
    # Create an optimizer
    optimizer_1 = non_linear.MultiNest()

    # Define galaxy priors
    source_galaxy_prior = galaxy_prior.GalaxyPrior(light_profile=light_profiles.EllipticalSersic,
                                                   variable_redshift=True)
    lens_galaxy_prior = galaxy_prior.GalaxyPrior(spherical_mass_profile=mass_profiles.EllipticalIsothermal,
                                                 shear_mass_profile=mass_profiles.ExternalShear, variable_redshift=True)

    # Add the galaxy priors to the optimizer
    optimizer_1.variable.source_galaxies = [source_galaxy_prior]
    optimizer_1.variable.lens_galaxies = [lens_galaxy_prior]

    optimizer_1.create_param_names()



from autolens.autopipe import model_mapper as mm
from autolens.analysis import galaxy_prior as gp
from autolens.profiles import light_profiles, mass_profiles
from autolens import config
import os


class TestCase:
    def test_integration(self):
        conf = config.DefaultPriorConfig(
            "{}/../{}".format(os.path.dirname(os.path.realpath(__file__)), "test_files/config/priors/default"))

        # Create a mapper. This can be used to convert values output by a non linear optimiser into class instances.
        mapper = mm.ModelMapper()

        # Create a galaxy prior for the source galaxy. Here we are describing only the light profile of the source
        # galaxy which comprises an elliptical exponential and elliptical sersic light profile.
        source_galaxy_prior = gp.GalaxyPrior(variable_redshift=True,
                                             light_profile_one=light_profiles.EllipticalExponential,
                                             light_profile_2=light_profiles.EllipticalSersic, config=conf)

        # Create a galaxy prior for the source galaxy. Here we are describing both the light and mass profiles. We've
        # also stipulated that the centres of any galaxies generated using the galaxy prior should match.
        lens_galaxy_prior = gp.GalaxyPrior(variable_redshift=True, light_profile=light_profiles.EllipticalExponential,
                                           mass_profile=mass_profiles.EllipticalExponentialMass,
                                           align_centres=True, config=conf)

        mapper.source_galaxy = source_galaxy_prior
        mapper.lens_galaxy = lens_galaxy_prior

        # Create a model instance. All the instances of the profile classes are created here. Normally we would do this
        # using the output of a non linear search but in this case we are using the median values from the priors.
        instance = mapper.instance_from_prior_medians()

        # Recover galaxy instances. We can pass the model instance to galaxy priors to recover a fully constructed
        # galaxy
        source_galaxy = instance.source_galaxy
        lens_galaxy = instance.lens_galaxy

        # Let's just check that worked
        assert len(source_galaxy.light_profiles) == 2
        assert len(source_galaxy.mass_profiles) == 0

        assert len(lens_galaxy.light_profiles) == 1
        assert len(lens_galaxy.mass_profiles) == 1

        assert source_galaxy.redshift == 1.5
        assert lens_galaxy.redshift == 1.5



from autolens.imaging.scaled_array import ScaledArray, AbstractArray
import numpy as np
from scipy.stats import norm
import scipy.signal
from autolens import exc


# TODO : The idea is that we don't need functions to estimate the noise or the exposure time once we set up an image
# TODO : so we'll leave this functionality to a class that loads images with what we're given.

class PrepatoryImage(ScaledArray):

    def __init__(self, array, pixel_scale, psf, noise=None, background_noise=None, poisson_noise=None,
                 effective_exposure_time=None):
        """
        A 2d array representing a real or simulated image.

        Parameters
        ----------
        array: ndarray
            An array of image pixels in gray-scale
        noise: ndarray
            An array describing the noise in the image
        effective_exposure_time: Union(ndarray, float)
            A float or array representing the effective exposure time of the whole image or each pixel.
        pixel_scale: float
            The scale of each pixel in arc seconds
        psf: PSF
            An array describing the PSF
        """
        super(PrepatoryImage, self).__init__(array, pixel_scale)
        self.psf = psf
        self.noise = noise
        self.background_noise = background_noise
        self.poisson_noise = poisson_noise
        self.effective_exposure_time = effective_exposure_time

    @classmethod
    def simulate(cls, array, pixel_scale, effective_exposure_time, psf=None, background_sky_map=None,
                 include_poisson_noise=False, seed=-1):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        array: ndarray
            A plain image
        effective_exposure_time: Union(ndarray, float)
            A float or array representing the effective exposure time of the whole image or each pixel.
        pixel_scale: float
            The scale of each pixel in arc seconds
        psf: PSF
            An array describing the PSF
        background_sky_map
        include_poisson_noise: Bool
            If True poisson noise is simulated and added to the image
        seed: int
            A seed for random noise generation

        Returns
        -------
        image: PrepatoryImage
            A simulated image
        """

        array_counts = None

        if background_sky_map is not None:
            array += background_sky_map
            background_noise_counts = np.sqrt(np.multiply(background_sky_map, effective_exposure_time))
            background_noise = np.divide(background_noise_counts, effective_exposure_time)
        else:
            background_noise_counts = None
            background_noise = None

        if psf is not None:
            array = psf.convolve(array)
            array = cls.trim_psf_edges(array, psf)
            effective_exposure_time = cls.trim_psf_edges(effective_exposure_time, psf)
            if background_sky_map is not None:
                background_sky_map = cls.trim_psf_edges(background_sky_map, psf)
            if background_noise_counts is not None:
                background_noise_counts = cls.trim_psf_edges(background_noise_counts, psf)
                background_noise = cls.trim_psf_edges(background_noise, psf)

        if include_poisson_noise is True:

            array += generate_poisson_noise(array, effective_exposure_time, seed)

            # The poisson noise map does not include the background sky, so this estimate below removes it
            if background_sky_map is not None:
                array_counts = np.multiply(np.subtract(array, background_sky_map), effective_exposure_time)
            elif background_sky_map is None:
                array_counts = np.multiply(array, effective_exposure_time)

            poisson_noise = np.divide(np.sqrt(array_counts), effective_exposure_time)

        else:

            poisson_noise = None

        # The final image is background subtracted.
        if background_sky_map is not None:
            array -= background_sky_map
        if background_sky_map is not None and include_poisson_noise is False:
            noise = np.divide(background_noise_counts, effective_exposure_time)
        elif background_sky_map is None and include_poisson_noise is True:
            noise = np.divide(array_counts, effective_exposure_time)
        elif background_sky_map is not None and include_poisson_noise is True:
            noise = np.divide(np.sqrt(array_counts + np.square(background_noise_counts)), effective_exposure_time)
        else:
            noise = None

        if noise is not None:
            if (np.isnan(noise)).any():
                raise exc.MaskException('Nan found in poisson noise - increase exposure time.')

        return PrepatoryImage(array, pixel_scale=pixel_scale, noise=noise, psf=psf, background_noise=background_noise,
                              poisson_noise=poisson_noise, effective_exposure_time=effective_exposure_time)

    def __array_finalize__(self, obj):
        super(PrepatoryImage, self).__array_finalize__(obj)
        if isinstance(obj, PrepatoryImage):
            self.psf = obj.psf
            self.noise = obj.noise
            self.background_noise = obj.background_noise
            self.poisson_noise = obj.poisson_noise
            self.effective_exposure_time = obj.effective_exposure_time

    @staticmethod
    def trim_psf_edges(array, psf):
        psf_cut_x = np.int(np.ceil(psf.shape[0] / 2)) - 1
        psf_cut_y = np.int(np.ceil(psf.shape[1] / 2)) - 1
        array_x = np.int(array.shape[0])
        array_y = np.int(array.shape[1])
        return array[psf_cut_x:array_x - psf_cut_x, psf_cut_y:array_y - psf_cut_y]

    def electrons_per_second_to_counts(self, array):
        """
        For an array (in electrons per second) and exposure time array, return an array in units counts.

        Parameters
        ----------
        array : ndarray
            The image from which the Poisson signal_to_noise_ratio map is estimated.
        """
        return np.multiply(array, self.effective_exposure_time)

    def counts_to_electrons_per_second(self, array):
        """
        For an array (in counts) and exposure time array, convert the array to units electrons per second

        Parameters
        ----------
        array : ndarray
            The image from which the Poisson signal_to_noise_ratio map is estimated.
        """
        return np.divide(array, self.effective_exposure_time)

    @property
    def counts_array(self):
        """
        Returns
        -------
        counts_array: ndarray
            An array representing the image in terms of counts
        """
        return self.electrons_per_second_to_counts(self)

    @property
    def background_noise_counts_array(self):
        """
        Returns
        -------
        background_noise_counts_array: ndarray
            An array representing the background noise in terms of counts
        """
        return self.electrons_per_second_to_counts(self.background_noise)

    @property
    def estimated_noise_counts(self):
        """
        Returns
        -------
        estimated_noise_counts: ndarray
            An array representing estimated noise in terms of counts
        """
        return np.sqrt(self.counts_array + np.square(self.background_noise_counts_array))

    @property
    def estimated_noise(self):
        """
        Returns
        -------
        estimated_noise: ndarray
            An array representing estimated noise
        """
        return self.counts_to_electrons_per_second(self.estimated_noise_counts)

    def background_noise_from_edges(self, no_edges):
        """Estimate the background signal_to_noise_ratio by binning data_to_image located at the edge(s) of an image
        into a histogram and fitting a Gaussian profiles to this histogram. The standard deviation (sigma) of this
        Gaussian gives a signal_to_noise_ratio estimate.

        Parameters
        ----------
        no_edges : int
            Number of edges used to estimate the background signal_to_noise_ratio.

        """

        edges = []

        for edge_no in range(no_edges):
            top_edge = self[edge_no, edge_no:self.shape[1] - edge_no]
            bottom_edge = self[self.shape[0] - 1 - edge_no, edge_no:self.shape[1] - edge_no]
            left_edge = self[edge_no + 1:self.shape[0] - 1 - edge_no, edge_no]
            right_edge = self[edge_no + 1:self.shape[0] - 1 - edge_no, self.shape[1] - 1 - edge_no]

            edges = np.concatenate((edges, top_edge, bottom_edge, right_edge, left_edge))

        return norm.fit(edges)[1]


class Image(ScaledArray):

    def __init__(self, array, pixel_scale, noise, psf):
        """
        A 2d array representing a real or simulated image.

        Parameters
        ----------
        array: ndarray
            An array of image pixels in gray-scale
        noise: ndarray
            An array describing the noise in the image
        pixel_scale: float
            The scale of each pixel in arc seconds
        psf: PSF
            An array describing the PSF
        """
        super(Image, self).__init__(array, pixel_scale)
        self.noise = noise
        self.psf = psf

    def __array_finalize__(self, obj):
        super(Image, self).__array_finalize__(obj)
        if isinstance(obj, Image):
            self.psf = obj.psf
            self.noise = obj.noise


class PSF(AbstractArray):

    # noinspection PyUnusedLocal
    def __init__(self, array, renormalize=True):
        """
        Class storing a 2D Point Spread Function (PSF), including its weighted_data and coordinate grid_coords.

        Parameters
        ----------
        array : ndarray
            The psf weighted_data.
        renormalize : bool
            Renormalize the PSF such that its value added up to 1.0?
        """

        # noinspection PyArgumentList
        super().__init__()
        if renormalize:
            self.renormalize()

    @classmethod
    def from_fits_renormalized(cls, file_path, hdu, pixel_scale):
        """
        Loads a PSF from fits and renormalizes it

        Parameters
        ----------
        file_path: String
            The path to the file containing the PSF
        hdu: int
            HDU ??
        pixel_scale: float
            The scale of a pixel in arcseconds

        Returns
        -------
        psf: PSF
            A renormalized PSF instance
        """
        psf = PSF.from_fits(file_path, hdu, pixel_scale)
        psf.renormalize()
        return psf

    def renormalize(self):
        """Renormalize the PSF such that its weighted_data values sum to unity."""
        return np.divide(self, np.sum(self))

    def convolve(self, array):
        """
        Convolve an array with this PSF

        Parameters
        ----------
        array: ndarray
            An array representing an image

        Returns
        -------
        convolved_array: ndarray
            An array representing an image that has been convolved with this PSF

        Raises
        ------
        KernelException if either PSF psf dimension is odd
        """
        if self.shape[0] % 2 == 0 or self.shape[1] % 2 == 0:
            raise exc.KernelException("PSF Kernel must be odd")

        return scipy.signal.convolve2d(array, self, mode='same')


def setup_random_seed(seed):
    """Setup the random seed. If the input seed is -1, the code will use a random seed for every run. If it is positive,
    that seed is used for all runs, thereby giving reproducible nlo

    Parameters
    ----------
    seed : int
        The seed of the random number generator, used for the random signal_to_noise_ratio maps.
    """
    if seed == -1:
        seed = np.random.randint(0,
                                 int(1e9))  # Use one seed, so all regions have identical column non-uniformity.
    np.random.seed(seed)


def generate_poisson_noise(image, exposure_time, seed=-1):
    """
    Generate a two-dimensional background noise-map for an image, generating values from a Gaussian
    distribution with mean 0.0.

    Parameters
    ----------
    image : ndarray
        The 2D image background noise is added to.
    exposure_time : Union(ndarray, int)
        The 2D array of pixel exposure times.
    seed : int
        The seed of the random number generator, used for the random noise maps.

    Returns
    -------
    poisson_noise: ndarray
        An array describing simulated poisson noise
    """
    setup_random_seed(seed)
    image_counts = np.multiply(image, exposure_time)
    return image - np.divide(np.random.poisson(image_counts, image.shape), exposure_time)



import numpy as np

from autolens.imaging import mask


class MockSubGridCoords(np.ndarray):
    def __new__(cls, sub_grid_coords, *args, **kwargs):
        return sub_grid_coords.view(cls)

    def __init__(self, sub_grid_coords, sub_to_image, sub_grid_size):
        # noinspection PyArgumentList
        super().__init__()
        self.sub_grid_coords = sub_grid_coords
        self.sub_to_image = sub_to_image
        self.no_pixels = sub_to_image.shape[0]
        self.sub_grid_size = sub_grid_size
        self.sub_grid_length = int(sub_grid_size ** 2.0)
        self.sub_grid_fraction = 1.0 / self.sub_grid_length


class MockGridCollection(object):

    def __init__(self, image, sub, blurring=None):
        self.image = mask.ImageGrid(image)
        self.sub = sub
        self.blurring = mask.ImageGrid(blurring) if blurring is not None else None


class MockBorderCollection(object):

    def __init__(self, image, sub):

        self.image = image
        self.sub = sub

    @classmethod
    def from_mask_and_subgrid_size(cls, mask, subgrid_size, polynomial_degree=3, centre=(0.0, 0.0)):
        image_border = ImageGridBorder.from_mask(mask, polynomial_degree, centre)
        sub_border = SubGridBorder.from_mask(mask, subgrid_size, polynomial_degree, centre)
        return BorderCollection(image_border, sub_border)

    def relocated_grids_from_grids(self, grids):
        return MockGridCollection(image=self.image.relocated_grid_from_grid(grids.image),
                              sub=MockSubGridCoords(self.sub.relocated_grid_from_grid(grids.sub), grids.sub.sub_to_image,
                                                    grids.sub.sub_grid_size),
                              blurring=None)



from autolens.imaging import image as im
from autolens.imaging import mask as msk
from autolens.imaging import convolution
from autolens.imaging import masked_image as mi
import numpy as np
import pytest


@pytest.fixture(name='image')
def make_image():
    psf = im.PSF(np.ones((1, 1)), 1)
    return im.Image(np.ones((3, 3)), pixel_scale=3., psf=psf, noise=np.ones((3, 3)))


@pytest.fixture(name="mask")
def make_mask():
    return msk.Mask(np.array([[True, False, True],
                              [False, False, False],
                              [True, False, True]]), pixel_scale=3.0)


@pytest.fixture(name="sparse_mask")
def make_sparse_mask(mask):
    return msk.SparseMask(mask, 1.0)


@pytest.fixture(name="masked_image")
def make_masked_image(image, mask):
    return mi.MaskedImage(image, mask)


class TestMaskedImage(object):
    def test_attributes(self, image, masked_image):
        assert image.pixel_scale == masked_image.pixel_scale
        assert image.psf == masked_image.psf

    def test_masking(self, masked_image):
        assert masked_image.noise.shape == (5,)

    def test_grids(self, masked_image):

        assert masked_image.grids.image.shape == (5, 2)

        assert (masked_image.grids.image == np.array([[-3, 0], [0, -3], [0, 0], [0, 3], [3, 0]])).all()
        assert (masked_image.grids.sub == np.array([[-3.5, -0.5], [-3.5, 0.5], [-2.5, -0.5], [-2.5, 0.5],
                                                    [-0.5, -3.5], [-0.5, -2.5], [0.5, -3.5], [0.5, -2.5],
                                                    [-0.5, -0.5], [-0.5, 0.5], [0.5, -0.5], [0.5, 0.5],
                                                    [-0.5, 2.5], [-0.5, 3.5], [0.5, 2.5], [0.5, 3.5],
                                                    [2.5, -0.5], [2.5, 0.5], [3.5, -0.5], [3.5, 0.5]])).all()

    def test_borders(self, masked_image):

        assert (masked_image.borders.image == np.array([0, 1, 2, 3, 4])).all()
        assert (masked_image.borders.sub == np.array([0, 4, 8, 13, 18])).all()

    def test_blurring_mask(self, masked_image):
        assert masked_image.blurring_mask.all()

    def test_convolvers(self, masked_image):
        assert type(masked_image.convolver_image) == convolution.ConvolverImage
        assert type(masked_image.convolver_mapping_matrix) == convolution.ConvolverMappingMatrix

    def test_map_to_2d(self, masked_image):
        assert (masked_image.map_to_2d(np.array([1, 1, 1, 1, 1])) == np.array([[0, 1, 0],
                                                                               [1, 1, 1],
                                                                               [0, 1, 0]])).all()

    def test_sparse_mask(self, mask, sparse_mask):
        assert (mask == sparse_mask).all()

    def test_subtract(self, masked_image):
        subtracted_image = masked_image - np.array([1, 0, 1, 0, 0])
        assert isinstance(subtracted_image, mi.MaskedImage)
        assert subtracted_image.psf == masked_image.psf
        assert subtracted_image.pixel_scale == masked_image.pixel_scale

        assert subtracted_image == np.array([0, 1, 0, 1, 1])



import numpy as np
from autolens.imaging import masked_image as mi
from autolens.imaging import mask
from autolens.analysis import ray_tracing

# TODO : Can we make model_image, galaxy_images, minimum_Values a part of hyper galaxies?

minimum_value_profile = 0.1


class Fitter(object):

    def __init__(self, masked_image, tracer):
        """
        Class to evaluate the fit between a model described by a tracer and an actual image.

        Parameters
        ----------
        masked_image: mi.MaskedImage
            An image that has been masked for efficiency
        tracer: ray_tracing.Tracer
            An object describing the model
        """
        self.image = masked_image
        self.tracer = tracer

    def fit_data_with_profiles_and_model_images(self, model_image, galaxy_images):
        """Fit the weighted_data using the ray_tracing model, where only light_profiles are used to represent the galaxy
        images.

        Parameters
        ----------
        model_image : ndarray
            The best-fit model image to the weighted_data, from a previous phase of the pipeline
        galaxy_images : [ndarray]
            The best-fit model image of each hyper-galaxy, which can tell us how much flux each pixel contributes
            to.
        """
        contributions = generate_contributions(model_image, galaxy_images, self.tracer.hyper_galaxies,
                                               [minimum_value_profile for _ in range(len(galaxy_images))])
        scaled_noise = self.scaled_noise_for_contributions(contributions)
        blurred_model_image = self.blurred_light_profile_image()
        fitness = compute_likelihood(self.image, scaled_noise, blurred_model_image)
        return fitness

    def scaled_noise_for_contributions(self, contributions):
        """Use the contributions of each hyper galaxy to compute the scaled noise.
        Parameters
        -----------
        contributions : [ndarray]
            The contribution of flux of each galaxy in each pixel (computed from galaxy.HyperGalaxy)
        """
        scaled_noises = list(
            map(lambda hyper, contribution: hyper.scaled_noise_for_contributions(self.image.noise, contribution),
                self.tracer.hyper_galaxies, contributions))
        return self.image.noise + sum(scaled_noises)

    def blurred_light_profile_image(self):
        """
        For a given ray-tracing model, compute the light profile image(s) of its galaxies and blur them with the
        PSF.
        """
        image_light_profile = self.tracer.generate_image_of_galaxy_light_profiles()
        blurring_image_light_profile = self.tracer.generate_blurring_image_of_galaxy_light_profiles()
        return blur_image_including_blurring_region(image_light_profile, blurring_image_light_profile,
                                                    self.image.convolver_image)

    def fit_data_with_profiles(self):
        """
        Fit the weighted_data using the ray_tracing model, where only light_profiles are used to represent the galaxy
        images.
        """
        blurred_model_image = self.blurred_light_profile_image()
        fitness = compute_likelihood(self.image, self.image.noise, blurred_model_image)
        return fitness


class PixelizedFitter(Fitter):
    def __init__(self, masked_image, sparse_mask, tracer):
        """
        Class to evaluate the fit between a model described by a tracer and an actual image.

        Parameters
        ----------
        masked_image: mi.MaskedImage
            An image that has been masked for efficiency
        sparse_mask: mask.SparseMask
            A mask describing which pixels should be used in clustering for pixelizations
        borders : mask.BorderCollection
            The pixels representing the border of each plane, used for relocation.
        tracer: ray_tracing.Tracer
            An object describing the model
        """
        super().__init__(masked_image, tracer)
        self.sparse_mask = sparse_mask

    def fit_data_with_pixelization(self):
        """Fit the weighted_data using the ray_tracing model, where only pixelizations are used to represent the galaxy
        images.
        """

        pix_pre_fit = self.tracer.reconstructors_from_source_plane(self.image.borders, self.sparse_mask)
        pix_fit = pix_pre_fit.reconstruct_image(self.image, self.image.noise,
                                                self.image.convolver_mapping_matrix)

        model_image = pix_fit.model_image_from_reconstruction_jit()

        return compute_pixelization_evidence(self.image, self.image.noise, model_image, pix_fit)

    def fit_data_with_pixelization_profiles_and_model_images(self, model_image, galaxy_images):
        raise NotImplementedError()

    def fit_data_with_pixelization_and_profiles(self):
        raise NotImplementedError()


def generate_contributions(model_image, galaxy_images, hyper_galaxies, minimum_values):
    """Use the model image and galaxy image (computed in the previous phase of the pipeline) to determine the
    contributions of each hyper galaxy.

    Parameters
    -----------
    minimum_values
    model_image : ndarray
        The best-fit model image to the weighted_data, from a previous phase of the pipeline
    galaxy_images : [ndarray]
        The best-fit model image of each hyper-galaxy, which can tell us how much flux each pixel contributes to.
    hyper_galaxies : [galaxy.HyperGalaxy]
        Each hyper-galaxy which is used to determine its contributions.
    """
    # noinspection PyArgumentList
    return list(map(lambda hyper, galaxy_image, minimum_value:
                    hyper.contributions_from_model_images(model_image, galaxy_image, minimum_value),
                    hyper_galaxies, galaxy_images, minimum_values))


def blur_image_including_blurring_region(image, blurring_image, convolver):
    """For a given image and blurring region, convert them to 2D and blur with the PSF, then return as
    the 1D DataGrid.

    Parameters
    ----------
    image : ndarray
        The image weighted_data using the GridData 1D representation.
    blurring_image : ndarray
        The blurring region weighted_data, using the GridData 1D representation.
    convolver : auto_lens.pixelization.frame_convolution.KernelConvolver
        The 2D Point Spread Function (PSF).
    """
    return convolver.convolve_image_jit(image, blurring_image)


def compute_likelihood(image, noise, model_image):
    """Compute the likelihood of a model image's fit to the weighted_data, by taking the difference between the
    observed image and model ray-tracing image. The likelihood consists of two terms:

    Chi-squared term - The residuals (model - weighted_data) of every pixel divided by the noise in each pixel, all
    squared.
    [Chi_Squared_Term] = sum(([Residuals] / [Noise]) ** 2.0)

    The overall normalization of the noise is also included, by summing the log noise value in each pixel:
    [Noise_Term] = sum(log(2*pi*[Noise]**2.0))

    These are summed and multiplied by -0.5 to give the likelihood:

    Likelihood = -0.5*[Chi_Squared_Term + Noise_Term]

    Parameters
    ----------
    image : grids.GridData
        The image weighted_data.
    noise : grids.GridData
        The noise in each pixel.
    model_image : grids.GridData
        The model image of the weighted_data.
    """
    return -0.5 * (compute_chi_sq_term(image, noise, model_image) + compute_noise_term(noise))


def compute_pixelization_evidence(image, noise, model_image, pix_fit):
    return -0.5 * (compute_chi_sq_term(image, noise, model_image)
                   + pix_fit.regularization_term_from_reconstruction()
                   + pix_fit.log_determinant_of_matrix_cholesky(pix_fit.covariance_regularization)
                   - pix_fit.log_determinant_of_matrix_cholesky(pix_fit.regularization)
                   + compute_noise_term(noise))


def compute_chi_sq_term(image, noise, model_image):
    """Compute the chi-squared of a model image's fit to the weighted_data, by taking the difference between the
    observed image and model ray-tracing image, dividing by the noise in each pixel and squaring:

    [Chi_Squared] = sum(([Data - Model] / [Noise]) ** 2.0)

    Parameters
    ----------
    image : grids.GridData
        The image weighted_data.
    noise : grids.GridData
        The noise in each pixel.
    model_image : grids.GridData
        The model image of the weighted_data.
    """
    return np.sum((np.add(image.view(float), - model_image) / noise) ** 2.0)


def compute_noise_term(noise):
    """Compute the noise normalization term of an image, which is computed by summing the noise in every pixel:

    [Noise_Term] = sum(log(2*pi*[Noise]**2.0))

    Parameters
    ----------
    noise : grids.GridData
        The noise in each pixel.
    """
    return np.sum(np.log(2 * np.pi * noise ** 2.0))


# noinspection PyUnusedLocal
def fit_data_with_pixelization_and_profiles(grid_data_collection, pixelization, convolver, tracer,
                                            mapper_cluster, image=None):
    # TODO: Implement me
    raise NotImplementedError("fit_data_with_pixelization_and_profiles has not been implemented")



from autolens.pipeline import pipeline as pl
from autolens.pipeline import phase as ph
from autolens.autopipe import model_mapper
from autolens.autopipe import non_linear
from autolens.imaging import image as im
from autolens.profiles import light_profiles
from autolens.profiles import mass_profiles
from autolens.analysis import galaxy_prior as gp
from autolens.analysis import galaxy as g
import numpy as np
import pytest


class DummyPhase(object):
    def __init__(self):
        self.masked_image = None
        self.previous_results = None
        self.name = "dummy_phase"

    def run(self, masked_image, previous_results):
        self.masked_image = masked_image
        self.previous_results = previous_results
        return non_linear.Result(model_mapper.ModelInstance(), 1)


class TestPipeline(object):
    def test_run_pipeline(self):
        phase_1 = DummyPhase()
        phase_2 = DummyPhase()
        pipeline = pl.Pipeline(phase_1, phase_2)

        pipeline.run(None)

        assert len(phase_1.previous_results) == 0
        assert len(phase_2.previous_results) == 1

    def test_addition(self):
        phase_1 = DummyPhase()
        phase_2 = DummyPhase()
        phase_3 = DummyPhase()

        pipeline1 = pl.Pipeline(phase_1, phase_2)
        pipeline2 = pl.Pipeline(phase_3)

        assert (phase_1, phase_2, phase_3) == (pipeline1 + pipeline2).phases


shape = (50, 50)


@pytest.fixture(name="profile_only_pipeline")
def make_profile_only_pipeline():
    return pl.make_profile_pipeline()


@pytest.fixture(name="image")
def make_image():
    return im.Image(np.ones(shape), pixel_scale=0.2, noise=np.ones(shape), psf=im.PSF(np.ones((3, 3))))


@pytest.fixture(name="results_1")
def make_results_1():
    const = model_mapper.ModelInstance()
    var = model_mapper.ModelMapper()
    const.lens_galaxy = g.Galaxy(elliptical_sersic=light_profiles.EllipticalSersic())
    var.lens_galaxy = gp.GalaxyPrior(elliptical_sersic=light_profiles.EllipticalSersic)
    return ph.SourceLensPhase.Result(const, 1, var, [np.full(shape, 0.5), None])


@pytest.fixture(name="results_2")
def make_results_2():
    const = model_mapper.ModelInstance()
    var = model_mapper.ModelMapper()
    var.lens_galaxy = gp.GalaxyPrior(sie=mass_profiles.SphericalIsothermal)
    var.source_galaxy = gp.GalaxyPrior(elliptical_sersic=light_profiles.EllipticalSersic)
    const.lens_galaxy = g.Galaxy(sie=mass_profiles.SphericalIsothermal())
    const.source_galaxy = g.Galaxy(elliptical_sersic=light_profiles.EllipticalSersic())
    return ph.SourceLensPhase.Result(const, 1, var, [np.full(shape, 0.5), np.full(shape, 0.5)])


@pytest.fixture(name="results_3")
def make_results_3():
    const = model_mapper.ModelInstance()
    var = model_mapper.ModelMapper()
    var.lens_galaxy = gp.GalaxyPrior(sie=mass_profiles.SphericalIsothermal,
                                     elliptical_sersic=light_profiles.EllipticalSersic)
    var.source_galaxy = gp.GalaxyPrior(elliptical_sersic=light_profiles.EllipticalSersic)
    const.lens_galaxy = g.Galaxy(sie=mass_profiles.SphericalIsothermal(),
                                 elliptical_sersic=light_profiles.EllipticalSersic())
    const.source_galaxy = g.Galaxy(elliptical_sersic=light_profiles.EllipticalSersic())
    return ph.SourceLensPhase.Result(const, 1, var, [np.full(shape, 0.5), np.full(shape, 0.5)])


@pytest.fixture(name="results_3h")
def make_results_3h():
    const = model_mapper.ModelInstance()
    var = model_mapper.ModelMapper()
    const.lens_galaxy = g.Galaxy(hyper_galaxy=g.HyperGalaxy())
    const.source_galaxy = g.Galaxy(hyper_galaxy=g.HyperGalaxy())
    return ph.SourceLensPhase.Result(const, 1, var, [np.full(shape, 0.5), np.full(shape, 0.5)])


class TestProfileOnlyPipeline(object):
    def test_phase1(self, profile_only_pipeline, image):
        phase1 = profile_only_pipeline.phases[0]
        analysis = phase1.make_analysis(image)

        assert isinstance(phase1.lens_galaxy, gp.GalaxyPrior)
        assert phase1.source_galaxy is None

        assert analysis.masked_image == np.ones((716,))
        assert analysis.masked_image.sub_grid_size == 1
        assert analysis.previous_results is None

    def test_phase2(self, profile_only_pipeline, image, results_1):
        phase2 = profile_only_pipeline.phases[1]
        previous_results = ph.ResultsCollection([results_1])
        analysis = phase2.make_analysis(image, previous_results)

        assert analysis.masked_image == np.full((704,), 0.5)

        assert isinstance(phase2.lens_galaxy, gp.GalaxyPrior)
        assert isinstance(phase2.source_galaxy, gp.GalaxyPrior)
        assert phase2.lens_galaxy.sie.centre == previous_results.first.variable.lens_galaxy.elliptical_sersic.centre

    def test_phase3(self, profile_only_pipeline, image, results_1, results_2):
        phase3 = profile_only_pipeline.phases[2]
        previous_results = ph.ResultsCollection([results_1, results_2])

        analysis = phase3.make_analysis(image, previous_results)

        assert isinstance(phase3.lens_galaxy, gp.GalaxyPrior)
        assert isinstance(phase3.source_galaxy, gp.GalaxyPrior)

        assert analysis.masked_image == np.ones((716,))

        assert phase3.lens_galaxy.elliptical_sersic == results_1.variable.lens_galaxy.elliptical_sersic
        assert phase3.lens_galaxy.sie == results_2.variable.lens_galaxy.sie
        assert phase3.source_galaxy == results_2.variable.source_galaxy

    def test_phase3h(self, profile_only_pipeline, image, results_1, results_2, results_3):
        phase3h = profile_only_pipeline.phases[3]
        previous_results = ph.ResultsCollection([results_1, results_2, results_3])

        analysis = phase3h.make_analysis(image, previous_results)

        assert isinstance(phase3h.lens_galaxy, gp.GalaxyPrior)
        assert isinstance(phase3h.source_galaxy, gp.GalaxyPrior)

        assert analysis.masked_image == np.ones((716,))

        assert phase3h.lens_galaxy.elliptical_sersic == results_3.constant.lens_galaxy.elliptical_sersic
        assert phase3h.lens_galaxy.sie == results_3.constant.lens_galaxy.sie
        assert phase3h.source_galaxy.elliptical_sersic == results_3.constant.source_galaxy.elliptical_sersic

        assert isinstance(phase3h.lens_galaxy.hyper_galaxy, model_mapper.PriorModel)
        assert isinstance(phase3h.source_galaxy.hyper_galaxy, model_mapper.PriorModel)

    def test_phase4(self, profile_only_pipeline, image, results_1, results_2, results_3, results_3h):
        phase4 = profile_only_pipeline.phases[4]
        previous_results = ph.ResultsCollection([results_1, results_2, results_3, results_3h])

        analysis = phase4.make_analysis(image, previous_results)

        assert isinstance(phase4.lens_galaxy, gp.GalaxyPrior)
        assert isinstance(phase4.source_galaxy, gp.GalaxyPrior)

        assert analysis.masked_image == np.ones((716,))

        assert isinstance(phase4.lens_galaxy.hyper_galaxy, g.HyperGalaxy)

        assert phase4.lens_galaxy.hyper_galaxy == results_3h.constant.lens_galaxy.hyper_galaxy
        assert phase4.source_galaxy.hyper_galaxy == results_3h.constant.source_galaxy.hyper_galaxy

        assert phase4.lens_galaxy.elliptical_sersic == results_3.variable.lens_galaxy.elliptical_sersic
        assert phase4.lens_galaxy.sie == results_3.variable.lens_galaxy.sie
        assert phase4.source_galaxy.elliptical_sersic == results_3.variable.source_galaxy.elliptical_sersic



import numpy as np
from autolens.imaging import convolution
from autolens.imaging import mask, scaled_array
import pytest
from autolens import exc


@pytest.fixture(name="simple_mask_index_array")
def make_simple_mask_index_array():
    return np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])


@pytest.fixture(name="cross_mask")
def make_cross_mask():
    mask = np.full((3, 3), False)

    mask[0, 0] = True
    mask[0, 2] = True
    mask[2, 2] = True
    mask[2, 0] = True

    return mask


@pytest.fixture(name="cross_mask_index_array")
def make_cross_mask_index_array():
    return np.array([[-1, 0, -1], [1, 2, 3], [-1, 4, -1]])


@pytest.fixture(name="simple_image_frame_indexes")
def make_simple_image_frame_indexes(simple_convolver):
    return simple_convolver.make_image_frame_indexes((3, 3))


@pytest.fixture(name="cross_image_frame_indexes")
def make_cross_image_frame_indexes(cross_convolver):
    return cross_convolver.make_image_frame_indexes((3, 3))


@pytest.fixture(name="cross_mask_image_frame_indexes")
def make_cross_mask_image_frame_indexes(cross_convolver):
    return cross_convolver.make_blurring_image_frame_indexes((3, 3), blurring_region_mask=np.full((3, 3), False))


@pytest.fixture(name="simple_convolver")
def make_simple_convolver():
    return convolution.ConvolverImage(mask=np.full((3, 3), False), blurring_mask=(np.full((3,3), False)),
                                      psf=np.array([[1.0, 2.0, 3.0],
                                                    [4.0, 5.0, 6.0],
                                                    [7.0, 8.0, 9.0]]))


@pytest.fixture(name="cross_convolver")
def make_cross_convolver(cross_mask):
    return convolution.ConvolverImage(mask=cross_mask, blurring_mask=(np.full((3,3), False)),
                                      psf=np.array([[1.0, 2.0, 3.0],
                                                    [4.0, 5.0, 6.0],
                                                    [7.0, 8.0, 9.0]]))


@pytest.fixture(name="simple_kernel")
def make_simple_kernel():
    return np.array([[0, 0.1, 0], [0.1, 0.6, 0.1], [0, 0.1, 0]])


class TestNumbering(object):

    def test_simple_numbering(self, simple_mask_index_array):

        shape = (3, 3)

        convolver = convolution.ConvolverImage(mask=np.full(shape, False), blurring_mask=np.full(shape, False),
                                               psf=np.ones((1, 1)))

        mask_index_array = convolver.mask_index_array

        assert mask_index_array.shape == shape
        # noinspection PyUnresolvedReferences
        assert (mask_index_array == simple_mask_index_array).all()

    def test_simple_mask(self, cross_mask):

        convolver = convolution.ConvolverImage(mask=cross_mask, blurring_mask=np.full(cross_mask.shape, False),
                                               psf=np.ones((1, 1)))

        assert (convolver.mask_index_array == np.array([[-1, 0, -1], [1, 2, 3], [-1, 4, -1]])).all()

    def test_even_failure(self):
        with pytest.raises(exc.KernelException):
            convolution.ConvolverImage(mask=np.full((3, 3), False), blurring_mask=np.full((3, 3), False),
                                       psf=np.ones((2, 2)))

    def test_mismatching_masks_failure(self):
        with pytest.raises(exc.KernelException):
            convolution.ConvolverImage(mask=np.full((3, 3), False),
                                       blurring_mask=np.full((3, 4), False), psf=np.ones((1, 1)))


class TestFrameExtraction(object):

    def test_trivial_frame_at_coords(self, simple_convolver):

        frame, psf_frame = simple_convolver.frame_at_coords_jit(coords=(1, 1), mask=np.full((3, 3), False),
                                                                   mask_index_array=simple_convolver.mask_index_array,
                                                                   psf=simple_convolver.psf)

        assert (frame == np.array([i for i in range(9)])).all()

    def test_corner_frame(self, simple_convolver):

        corner_frame =  np.array([0, 1, 3, 4, -1, -1, -1, -1, -1])

        frame, psf_frame = simple_convolver.frame_at_coords_jit(coords=(0, 0), mask=np.full((3, 3), False),
                                                                   mask_index_array=simple_convolver.mask_index_array,
                                                                   psf=simple_convolver.psf)

        assert (frame == corner_frame).all()

    def test_trivial_psf_frame_at_coords(self, simple_convolver):

        frame, psf_frame = simple_convolver.frame_at_coords_jit(coords=(1, 1), mask=np.full((3, 3), False),
                                                                   mask_index_array=simple_convolver.mask_index_array,
                                                                   psf=simple_convolver.psf)

        assert (psf_frame == np.array([[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])).all()

    def test_corner_psf_frame(self, simple_convolver):

        corner_frame =  [0, 1, 3, 4]

        frame, psf_frame = simple_convolver.frame_at_coords_jit(coords=(0, 0), mask=np.full((3, 3), False),
                                                                   mask_index_array=simple_convolver.mask_index_array,
                                                                   psf=simple_convolver.psf)

        assert (psf_frame == np.array([5.0, 6.0, 8.0, 9.0, -1, -1, -1, -1, -1])).all()

    def test_simple_square(self, simple_convolver):

        assert 9 == len(simple_convolver.image_frame_indexes)

        assert (simple_convolver.image_frame_indexes[4] == np.array([i for i in range(9)])).all()

    def test_frame_5x5_kernel__at_coords(self):

        convolver = convolution.ConvolverImage(mask=np.full((3, 3), False), blurring_mask=(np.full((3, 3), False)),
                                               psf=np.array([[1.0, 2.0, 3.0, 4.0, 5.0],
                                                             [6.0,  7.0,  8.0,  9.0,  10.0],
                                                             [11.0, 12.0, 13.0, 14.0, 15.0],
                                                             [16.0, 17.0, 18.0, 19.0, 20.0],
                                                             [21.0, 22.0, 23.0, 24.0, 25.0]]))

        frame, psf_frame = convolver.frame_at_coords_jit(coords=(0, 0), mask=np.full((3, 3), False),
                                                            mask_index_array=convolver.mask_index_array,
                                                            psf=convolver.psf)

        assert (psf_frame == np.array([13.0, 14.0, 15.0, 18.0, 19.0, 20.0, 23.0, 24.0, 25.0, -1., -1., -1., -1., -1.,
                                          -1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1.])).all()

        frame, psf_frame = convolver.frame_at_coords_jit(coords=(1, 0), mask=np.full((3, 3), False),
                                                            mask_index_array=convolver.mask_index_array,
                                                            psf=convolver.psf)

        assert (psf_frame == np.array([8.0, 9.0, 10.0, 13.0, 14.0, 15.0, 18.0, 19.0, 20.0, -1., -1., -1., -1., -1.,
                                          -1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1])).all()

        frame, psf_frame = convolver.frame_at_coords_jit(coords=(1, 1), mask=np.full((3, 3), False),
                                                            mask_index_array=convolver.mask_index_array,
                                                            psf=convolver.psf)

        assert (psf_frame == np.array([7.0, 8.0, 9.0, 12.0, 13.0, 14.0, 17.0, 18.0, 19.0, -1., -1., -1., -1., -1.,
                                          -1., -1., -1., -1., -1., -1., -1., -1., -1., -1., -1])).all()


class TestImageFrameIndexes(object):

    def test_masked_cross__3x3_kernel(self, cross_convolver):

        assert 5 == len(cross_convolver.image_frame_indexes)

        assert (cross_convolver.image_frame_indexes[0] == np.array([0, 1, 2, 3, -1, -1, -1, -1, -1])).all()
        assert (cross_convolver.image_frame_indexes[1] == np.array([0, 1, 2, 4, -1, -1, -1, -1, -1])).all()
        assert (cross_convolver.image_frame_indexes[2] == np.array([0, 1, 2, 3, 4, -1, -1, -1, -1])).all()
        assert (cross_convolver.image_frame_indexes[3] == np.array([0, 2, 3, 4, -1, -1, -1, -1, -1])).all()
        assert (cross_convolver.image_frame_indexes[4] == np.array([1, 2, 3, 4, -1, -1, -1, -1, -1])).all()

    def test_masked_square__3x5_kernel__loses_edge_of_top_and_bottom_rows(self):

        convolver = convolution.ConvolverImage(mask=np.full((3, 3), False), blurring_mask=(np.full((3, 3), False)),
                                               psf=np.ones((3, 5)))

        assert (convolver.image_frame_indexes[0] == np.array([0, 1, 2, 3, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[1] == np.array([0, 1, 2, 3, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[2] == np.array([0, 1, 2, 3, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[3] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[4] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[5] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[6] == np.array([3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[7] == np.array([3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[8] == np.array([3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()

    def test_masked_square__5x3_kernel__loses_edge_of_left_and_right_columns(self):

        convolver = convolution.ConvolverImage(mask=np.full((3, 3), False), blurring_mask=(np.full((3, 3), False)),
                                               psf=np.ones((5, 3)))

        assert (convolver.image_frame_indexes[0] == np.array([0, 1, 3, 4, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[1] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[2] == np.array([1, 2, 4, 5, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[3] == np.array([0, 1, 3, 4, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[4] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[5] == np.array([1, 2, 4, 5, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[6] == np.array([0, 1, 3, 4, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[7] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[8] == np.array([1, 2, 4, 5, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()

    def test_masked_square__5x5_kernel(self):

        convolver = convolution.ConvolverImage(mask=np.full((3, 3), False), blurring_mask=(np.full((3, 3), False)),
                                               psf=np.ones((5, 5)))

        assert (convolver.image_frame_indexes[0] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[1] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[2] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[3] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[4] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[5] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[6] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[7] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_indexes[8] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1,-1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()


class TestImageFrameKernels(object):

    def test_simple_square(self, simple_convolver):

        assert 9 == len(simple_convolver.image_frame_indexes)

        assert (simple_convolver.image_frame_psfs[4] == np.array([[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])).all()

    def test_masked_square__3x5_kernel__loses_edge_of_top_and_bottom_rows(self):

        convolver = convolution.ConvolverImage(mask=np.full((3, 3), False), blurring_mask=(np.full((3, 3), False)),
                                               psf=np.array([[1.0, 2.0, 3.0, 4.0, 5.0],
                                                             [6.0, 7.0, 8.0, 9.0, 10.0],
                                                             [11.0, 12.0, 13.0, 14.0, 15.0]]))

        assert (convolver.image_frame_psfs[0] == np.array([8.0, 9.0, 10.0, 13.0, 14.0, 15.0, -1, -1, -1,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[1] == np.array([7.0, 8.0, 9.0, 12.0, 13.0, 14.0, -1, -1, -1,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[2] == np.array([6.0, 7.0, 8.0, 11.0, 12.0, 13.0, -1, -1, -1,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[3] == np.array([3.0, 4.0, 5.0, 8.0, 9.0, 10.0, 13.0, 14.0, 15.0,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[4] == np.array([2.0, 3.0, 4.0, 7.0, 8.0, 9.0, 12.0, 13.0, 14.0,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[5] == np.array([1.0, 2.0, 3.0, 6.0, 7.0, 8.0, 11.0, 12.0, 13.0,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[6] == np.array([3.0, 4.0, 5.0, 8.0, 9.0, 10.0, -1, -1, -1,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[7] == np.array([2.0, 3.0, 4.0, 7.0, 8.0, 9.0, -1, -1, -1,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[8] == np.array([1.0, 2.0, 3.0, 6.0, 7.0, 8.0, -1, -1, -1,
                                                           -1, -1, -1, -1, -1, -1])).all()

    def test_masked_square__5x3_kernel__loses_edge_of_left_and_right_columns(self):

        convolver = convolution.ConvolverImage(mask=np.full((3, 3), False), blurring_mask=(np.full((3, 3), False)),
                                               psf=np.array([[1.0, 2.0, 3.0],
                                                             [4.0, 5.0, 6.0],
                                                             [7.0, 8.0, 9.0],
                                                             [10.0, 11.0, 12.0],
                                                             [13.0, 14.0, 15.0]]))

        assert (convolver.image_frame_psfs[0] == np.array([8.0, 9.0, 11.0, 12.0, 14.0, 15.0, -1, -1, -1,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[1] == np.array([7.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0, 14.0, 15.0,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[2] == np.array([7.0, 8.0, 10.0, 11.0, 13.0, 14.0, -1, -1, -1,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[3] == np.array([5.0, 6.0, 8.0, 9.0, 11.0, 12.0, -1, -1, -1,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[4] == np.array([4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[5] == np.array([4.0, 5.0, 7.0, 8.0, 10.0, 11.0, -1, -1, -1,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[6] == np.array([2.0, 3.0, 5.0, 6.0, 8.0, 9.0, -1, -1, -1,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[7] == np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0,
                                                           -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.image_frame_psfs[8] == np.array([1.0, 2.0, 4.0, 5.0, 7.0, 8.0, -1, -1, -1,
                                                           -1, -1, -1, -1, -1, -1])).all()


class TestBlurringFrameIndxes(object):

    def test__blurring_region_3x3_kernel(self, cross_mask):

        blurring_mask = np.array([[False, True, False],
                                  [True, True, True],
                                  [False, True, False]])

        convolver = convolution.ConvolverImage(mask=cross_mask, blurring_mask=blurring_mask, psf=np.ones((3, 3)))

        assert (convolver.blurring_frame_indexes[0] == np.array([0, 1, 2, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.blurring_frame_indexes[1] == np.array([0, 2, 3, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.blurring_frame_indexes[2] == np.array([1, 2, 4, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.blurring_frame_indexes[3] == np.array([2, 3, 4, -1, -1, -1, -1, -1, -1])).all()

    def test__blurring_region_5x5_kernel(self, cross_mask):

        blurring_mask = np.array([[False, True, False],
                                  [True, True, True],
                                  [False, True, False]])

        convolver = convolution.ConvolverImage(mask=cross_mask, blurring_mask=blurring_mask, psf=np.ones((5, 5)))

        assert (convolver.blurring_frame_indexes[0] == np.array([0, 1, 2, 3, 4, -1, -1, -1, -1, -1, -1,
                                        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.blurring_frame_indexes[1] == np.array([0, 1, 2, 3, 4, -1, -1, -1, -1, -1, -1,
                                        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.blurring_frame_indexes[2] == np.array([0, 1, 2, 3, 4, -1, -1, -1, -1, -1, -1,
                                        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.blurring_frame_indexes[3] == np.array([0, 1, 2, 3, 4, -1, -1, -1, -1, -1, -1,
                                        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()


class TestBlurringFrameKernels(object):

    def test__blurring_region_3x3_kernel(self, cross_mask):

        blurring_mask = np.array([[False, True, False],
                                  [True, True, True],
                                  [False, True, False]])


        convolver = convolution.ConvolverImage(mask=cross_mask, blurring_mask=blurring_mask,
                                               psf=np.array([[1.0, 2.0, 3.0],
                                                             [4.0, 5.0, 6.0],
                                                             [7.0, 8.0, 9.0]]))

        assert (convolver.blurring_frame_psfs[0] == np.array([6.0, 8.0, 9.0, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.blurring_frame_psfs[1] == np.array([4.0, 7.0, 8.0, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.blurring_frame_psfs[2] == np.array([2.0, 3.0, 6.0, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.blurring_frame_psfs[3] == np.array([1.0, 2.0, 4.0, -1, -1, -1, -1, -1, -1])).all()

    def test__blurring_region_5x5_kernel(self, cross_mask):

        blurring_mask = np.array([[False, True, False],
                                  [True, True, True],
                                  [False, True, False]])


        convolver = convolution.ConvolverImage(mask=cross_mask, blurring_mask=blurring_mask,
                                               psf=np.array([[1.0, 2.0, 3.0, 4.0, 5.0],
                                                             [6.0, 7.0, 8.0, 9.0, 10.0],
                                                             [11.0, 12.0, 13.0, 14.0, 15.0],
                                                             [16.0, 17.0, 18.0, 19.0, 20.0],
                                                             [21.0, 22.0, 23.0, 24.0, 25.0]]))

        assert (convolver.blurring_frame_psfs[0] == np.array([14.0, 18.0, 19.0, 20.0, 24.0, -1, -1, -1, -1, -1, -1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.blurring_frame_psfs[1] == np.array([12.0, 16.0, 17.0, 18.0, 22.0, -1, -1, -1, -1, -1, -1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.blurring_frame_psfs[2] == np.array([4.0, 8.0, 9.0, 10.0, 14.0, -1, -1, -1, -1, -1, -1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()
        assert (convolver.blurring_frame_psfs[3] == np.array([2.0, 6.0, 7.0, 8.0, 12.0, -1, -1, -1, -1, -1, -1,
                                                              -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])).all()


class TestFrameLengths(object):

    def test__frames_are_from_examples_above__lengths_are_right(self):

        convolver = convolution.ConvolverImage(mask=np.full((3, 3), False), blurring_mask=(np.full((3, 3), False)),
                                               psf=np.ones((3, 5)))

        # convolver_image.image_frame_indexes[0] == np.array([0, 1, 2, 3, 4, 5])
        # convolver_image.image_frame_indexes[1] == np.array([0, 1, 2, 3, 4, 5])
        # convolver_image.image_frame_indexes[2] == np.array([0, 1, 2, 3, 4, 5])
        # convolver_image.image_frame_indexes[3] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        # (convolver_image.image_frame_indexes[4] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        # convolver_image.image_frame_indexes[5] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        # convolver_image.image_frame_indexes[6] == np.array([3, 4, 5, 6, 7, 8])
        # convolver_image.image_frame_indexes[7] == np.array([3, 4, 5, 6, 7, 8])
        # convolver_image.image_frame_indexes[8] == np.array([3, 4, 5, 6, 7, 8])

        assert (convolver.image_frame_lengths == np.array([6, 6, 6, 9 , 9 , 9, 6, 6, 6])).all()

    def test__blurring_frames_from_example_above__lengths_are_right(self, cross_mask):

        blurring_mask = np.array([[False, True, False],
                                  [True, True, True],
                                  [False, True, False]])


        convolver = convolution.ConvolverImage(mask=cross_mask, blurring_mask=blurring_mask,
                                               psf=np.array([[1.0, 2.0, 3.0, 4.0, 5.0],
                                                             [6.0, 7.0, 8.0, 9.0, 10.0],
                                                             [11.0, 12.0, 13.0, 14.0, 15.0],
                                                             [16.0, 17.0, 18.0, 19.0, 20.0],
                                                             [21.0, 22.0, 23.0, 24.0, 25.0]]))

        # assert (convolver_image.blurring_frame_psfs[0] == np.array([14.0, 18.0, 19.0, 20.0, 24.0])).all()
        # assert (convolver_image.blurring_frame_psfs[1] == np.array([12.0, 16.0, 17.0, 18.0, 22.0])).all()
        # assert (convolver_image.blurring_frame_psfs[2] == np.array([4.0, 8.0, 9.0, 10.0, 14.0])).all()
        # assert (convolver_image.blurring_frame_psfs[3] == np.array([2.0, 6.0, 7.0, 8.0, 12.0])).all()

        assert (convolver.blurring_frame_lengths == np.array([5, 5, 5, 5])).all()


@pytest.fixture(name="convolver_4_edges")
def make_convolver_4_edges():
    mask = np.array(
        [[True, True, True, True],
         [True, False, False, True],
         [True, False, False, True],
         [True, True, True, True]]
    )

    convolver = convolution.ConvolverImage(mask)
    return convolver.convolver_for_kernel_shape((3, 3), mask)


class TestConvolution(object):

    def test_cross_mask_with_blurring_entries(self, cross_mask):

        kernel = np.array([[0, 0.2, 0],
                           [0.2, 0.4, 0.2],
                           [0, 0.2, 0]])

        blurring_mask = np.array([[False, True, False],
                                  [True, True, True],
                                  [False, True, False]])

        convolver = convolution.ConvolverImage(mask=cross_mask, blurring_mask=blurring_mask, psf=kernel)

        pixel_array = np.array([1, 0, 0, 0, 0])
        blurring_array = np.array([1, 0, 0, 0])

        result = convolver.convolve_image_jit(pixel_array, blurring_array)

        assert (np.round(result, 1) == np.array([0.6, 0.2, 0.2, 0., 0.])).all()


class TestConvolveMappingMatrix(object):

    def test__asymetric_convolver__matrix_blurred_correctly(self):

        shape = (4, 4)
        mask = np.full(shape, False)

        asymmetric_kernel = np.array([[0, 0.0, 0],
                                      [0.4, 0.2, 0.3],
                                      [0, 0.1, 0]])

        convolver = convolution.ConvolverMappingMatrix(mask=mask, psf=asymmetric_kernel)

        mapping = np.array([[0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0 ,0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 1, 0], # The 0.3 should be 'chopped' from this pixel as it is on the right-most edge
                            [0, 0, 0],
                            [1, 0, 0],
                            [0, 0, 1],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0]])

        blurred_mapping = convolver.convolve_mapping_matrix_jit(mapping)

        assert (blurred_mapping == np.array([[0,     0,   0],
                                             [0,     0,   0],
                                             [0,     0,   0],
                                             [0,     0,   0],
                                             [0,     0,   0],
                                             [0,     0,   0],
                                             [0,   0.4,   0],
                                             [0,   0.2,   0],
                                             [0.4,   0,   0],
                                             [0.2,   0, 0.4],
                                             [0.3,   0, 0.2],
                                             [0,   0.1, 0.3],
                                             [0,     0,   0],
                                             [0.1,   0,   0],
                                             [0,     0, 0.1],
                                             [0,     0,   0]])).all()

    def test__asymetric_convolver__multiple_overlapping_blurred_entires_in_matrix(self):

        shape = (4, 4)
        mask = np.full(shape, False)

        asymmetric_kernel = np.array([[0, 0.0, 0],
                                      [0.4, 0.2, 0.3],
                                      [0, 0.1, 0]])

        convolver = convolution.ConvolverMappingMatrix(mask=mask, psf=asymmetric_kernel)

        mapping = np.array([[0, 1, 0],
                            [0, 1, 0],
                            [0, 1, 0],
                            [0, 0, 0],
                            [0, 0 ,0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 1, 0], # The 0.3 should be 'chopped' from this pixel as it is on the right-most edge
                            [1, 0, 0],
                            [1, 0, 0],
                            [0, 0, 1],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0]])

        blurred_mapping = convolver.convolve_mapping_matrix_jit(mapping)

        assert blurred_mapping == pytest.approx(np.array([[0,     0.6,   0],
                                                          [0,     0.9,   0],
                                                          [0,     0.5,   0],
                                                          [0,     0.3,   0],
                                                          [0,     0.1,   0],
                                                          [0,     0.1,   0],
                                                          [0,   0.5,   0],
                                                          [0,   0.2,   0],
                                                          [0.6,   0,   0],
                                                          [0.5,   0, 0.4],
                                                          [0.3,   0, 0.2],
                                                          [0,   0.1, 0.3],
                                                          [0.1,   0,   0],
                                                          [0.1,   0,   0],
                                                          [0,     0, 0.1],
                                                          [0,     0,   0]]), 1e-4)